import pytest
import numpy as np
from scipy import ndimage
from libertem import masks
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random

from libertem.analysis.com import apply_correction, guess_corrections


@pytest.fixture
def ds_w_zero_frame():
    data = _mk_random(size=(16, 16, 16, 16))
    data[0, 0] = np.zeros((16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2,
    )
    return dataset


@pytest.fixture
def ds_random():
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16, 16),
        num_partitions=2,
    )
    return dataset


def test_com_with_zero_frames(ds_w_zero_frame, lt_ctx):
    analysis = lt_ctx.create_com_analysis(
        dataset=ds_w_zero_frame, cx=0, cy=0, mask_radius=0
    )
    results = lt_ctx.run(analysis)

    # no inf/nan in center_x and center_y
    assert not np.any(np.isinf(results[3].raw_data))
    assert not np.any(np.isinf(results[4].raw_data))
    assert not np.any(np.isnan(results[3].raw_data))
    assert not np.any(np.isnan(results[4].raw_data))

    # no inf/nan in divergence/magnitude
    assert not np.any(np.isinf(results[1].raw_data))
    assert not np.any(np.isinf(results[2].raw_data))
    assert not np.any(np.isnan(results[1].raw_data))
    assert not np.any(np.isnan(results[2].raw_data))


def test_com_comparison_scipy_1_nomask(ds_random, lt_ctx):
    analysis = lt_ctx.create_com_analysis(
        dataset=ds_random, cx=0, cy=0, mask_radius=None
    )
    results = lt_ctx.run(analysis)
    raw_data_by_frame = ds_random.data.reshape((16 * 16, 16, 16))
    field_x, field_y = results.field.raw_data
    field_x = field_x.reshape(16 * 16)
    field_y = field_y.reshape(16 * 16)
    for idx in range(16 * 16):
        scy, scx = ndimage.center_of_mass(raw_data_by_frame[idx])
        assert np.allclose(scx, field_x[idx])
        assert np.allclose(scy, field_y[idx])


def test_com_comparison_scipy_2_masked(ds_random, lt_ctx):
    analysis = lt_ctx.create_com_analysis(
        dataset=ds_random, cx=0, cy=0, mask_radius=8
    )
    results = lt_ctx.run(analysis)
    raw_data_by_frame = ds_random.data.reshape((16 * 16, 16, 16))
    field_x, field_y = results.field.raw_data
    field_x = field_x.reshape(16 * 16)
    field_y = field_y.reshape(16 * 16)
    disk_mask = masks.circular(
        centerX=0, centerY=0,
        imageSizeX=16,
        imageSizeY=16,
        radius=8,
    )
    for idx in range(16 * 16):
        masked_frame = raw_data_by_frame[idx] * disk_mask
        scy, scx = ndimage.center_of_mass(masked_frame)
        assert np.allclose(scx, field_x[idx])
        assert np.allclose(scy, field_y[idx])


def test_com_fails_with_non_4d_data_1(lt_ctx):
    data = _mk_random(size=(16 * 16, 16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16, 16),
        num_partitions=32,
    )
    with pytest.raises(Exception):
        lt_ctx.create_com_analysis(
            dataset=dataset, cx=0, cy=0, mask_radius=8
        )


def test_com_fails_with_non_4d_data_2(lt_ctx):
    data = _mk_random(size=(16, 16, 16 * 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16 * 16),
        num_partitions=16,
        sig_dims=1,
    )
    with pytest.raises(Exception):
        lt_ctx.create_com_analysis(
            dataset=dataset, cx=0, cy=0, mask_radius=8
        )


def test_com_linescan(lt_ctx):
    data = _mk_random(size=(1, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16, 16),
        num_partitions=4,
    )
    analysis = lt_ctx.create_com_analysis(
        dataset=dataset, cx=0, cy=0, mask_radius=8
    )
    lt_ctx.run(analysis)


def test_com_complex_numbers(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="complex64")
    ds_complex = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
    )
    analysis = lt_ctx.create_com_analysis(dataset=ds_complex, cx=0, cy=0, mask_radius=None)
    results = lt_ctx.run(analysis)

    reshaped_data = ds_complex.data.reshape((16 * 16, 16, 16))
    field_x = results.x_real.raw_data + 1j * results.x_imag.raw_data
    field_y = results.y_real.raw_data + 1j * results.y_imag.raw_data

    field_x = field_x.reshape(16 * 16)
    field_y = field_y.reshape(16 * 16)
    for idx in range(16 * 16):
        scy, scx = ndimage.center_of_mass(reshaped_data[idx])

        print(scx, field_x[idx])

        # difference between scipy and our impl: we don't divide by zero
        if np.isinf(scx):
            assert field_x[idx] == 0
            assert field_y[idx] == 0
        else:
            assert np.allclose(scx, field_x[idx])
            assert np.allclose(scy, field_y[idx])


def test_com_complex_numbers_handcrafted_1(lt_ctx):
    data = np.ones((3, 3, 4, 4), dtype="complex64")
    data[0, 0] = np.array([
        [0,    0,    0, 0],
        [0, 1+2j, 1-2j, 0],
        [0, 1-2j, 1+2j, 0],
        [0,    0,    0, 0],
    ])
    ds_complex = MemoryDataSet(
        data=data,
        tileshape=(1, 4, 4),
        num_partitions=9,
    )
    analysis = lt_ctx.create_com_analysis(dataset=ds_complex, cx=0, cy=0, mask_radius=None)
    results = lt_ctx.run(analysis)

    field_x = results.x_real.raw_data + 1j * results.x_imag.raw_data
    field_y = results.y_real.raw_data + 1j * results.y_imag.raw_data

    assert field_x[0, 0] == 1.5
    assert field_y[0, 0] == 1.5


def test_com_complex_numbers_handcrafted_2(lt_ctx):
    data = np.ones((3, 3, 4, 4), dtype="complex64")
    data[0, 0] = np.array([
        [0,    0,    0, 0],
        [0,    0, 1-2j, 0],
        [0, 1-2j,    0, 0],
        [0,    0,    0, 0],
    ])
    ds_complex = MemoryDataSet(
        data=data,
        tileshape=(1, 4, 4),
        num_partitions=9,
    )
    analysis = lt_ctx.create_com_analysis(dataset=ds_complex, cx=0, cy=0, mask_radius=None)
    results = lt_ctx.run(analysis)

    field_x = results.x_real.raw_data + 1j * results.x_imag.raw_data
    field_y = results.y_real.raw_data + 1j * results.y_imag.raw_data

    assert field_x[0, 0] == 1.5
    assert field_y[0, 0] == 1.5


def test_com_complex_numbers_handcrafted_3(lt_ctx):
    data = np.ones((3, 3, 4, 4), dtype="complex64")
    data[0, 0] = np.array([
        0,    0,    0, 0,
        0,    0, 1-2j, 0,
        0,    0,    0, 0,
        0,    0,    0, 0,
    ], dtype="complex64").reshape((4, 4))
    ds_complex = MemoryDataSet(
        data=data,
        tileshape=(1, 4, 4),
        num_partitions=9,
    )
    analysis = lt_ctx.create_com_analysis(dataset=ds_complex, cx=0, cy=0, mask_radius=None)
    results = lt_ctx.run(analysis)

    print(data[0, 0])

    field_x = results.x_real.raw_data + 1j * results.x_imag.raw_data
    field_y = results.y_real.raw_data + 1j * results.y_imag.raw_data

    assert field_x[0, 0] == 2
    assert field_y[0, 0] == 1


def test_com_default_params(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16, 16),
        num_partitions=16,
        sig_dims=2,
    )
    analysis = lt_ctx.create_com_analysis(
        dataset=dataset,
    )
    lt_ctx.run(analysis)


def test_com_divergence(lt_ctx):
    data = np.zeros((3, 3, 3, 3), dtype=np.float32)
    for i in range(3):
        for j in range(3):
            data[i, j, i, j] = 1
    dataset = MemoryDataSet(
        data=data,
        sig_dims=2,
    )
    analysis = lt_ctx.create_com_analysis(
        dataset=dataset,
        cy=1,
        cx=1,

    )
    res = lt_ctx.run(analysis)

    print(data)
    print("y", res["y"].raw_data)
    print("x", res["x"].raw_data)
    print("divergence", res["divergence"].raw_data)
    print("curl", res["curl"].raw_data)
    assert np.all(res["x"].raw_data == [[-1, 0, 1], [-1, 0, 1], [-1, 0, 1]])
    assert np.all(res["y"].raw_data == [[-1, -1, -1], [0, 0, 0], [1, 1, 1]])

    assert np.all(res["divergence"].raw_data == 2)
    assert np.all(res["curl"].raw_data == 0)


def test_com_divergence_2(lt_ctx):
    data = np.zeros((3, 3, 3, 3), dtype=np.float32)
    for i in range(3):
        for j in range(3):
            data[i, j, 2-i, 2-j] = 1
    dataset = MemoryDataSet(
        data=data,
        sig_dims=2,
    )
    analysis = lt_ctx.create_com_analysis(
        dataset=dataset,
        cy=1,
        cx=1,
    )
    res = lt_ctx.run(analysis)

    print(data)
    print("y", res["y"].raw_data)
    print("x", res["x"].raw_data)
    print("divergence", res["divergence"].raw_data)
    print("curl", res["curl"].raw_data)
    assert np.all(res["x"].raw_data == [[1, 0, -1], [1, 0, -1], [1, 0, -1]])
    assert np.all(res["y"].raw_data == [[1, 1, 1], [0, 0, 0], [-1, -1, -1]])

    assert np.all(res["divergence"].raw_data == -2)
    assert np.all(res["curl"].raw_data == 0)


def test_com_curl(lt_ctx):
    data = np.zeros((3, 3, 3, 3), dtype=np.float32)
    for y in range(3):
        for x in range(3):
            data[y, x, x, 2-y] = 1
    dataset = MemoryDataSet(
        data=data,
        sig_dims=2,
    )
    analysis = lt_ctx.create_com_analysis(
        dataset=dataset,
        cy=1,
        cx=1
    )
    res = lt_ctx.run(analysis)

    print(data)
    print("y", res["y"].raw_data)
    print("x", res["x"].raw_data)
    print("divergence", res["divergence"].raw_data)
    print("curl", res["curl"].raw_data)

    assert np.all(res["x"].raw_data == [[1, 1, 1], [0, 0, 0], [-1, -1, -1]])
    assert np.all(res["y"].raw_data == [[-1, 0, 1], [-1, 0, 1], [-1, 0, 1]])

    assert np.all(res["divergence"].raw_data == 0)
    # Data "rotating to the right"
    # Y points down, x points right, right-handed coordinate system: z points away
    # Right handed coordinate system: The result should be positive
    # since the patterns are rotated in positive direction
    # https://commons.wikimedia.org/wiki/File:Right_hand_rule_simple.png
    assert np.all(res["curl"].raw_data == 2)


def test_com_curl_2(lt_ctx):
    data = np.zeros((3, 3, 3, 3), dtype=np.float32)
    for y in range(3):
        for x in range(3):
            data[y, x, 2-x, y] = 1
    dataset = MemoryDataSet(
        data=data,
        sig_dims=2,
    )
    analysis = lt_ctx.create_com_analysis(
        dataset=dataset,
        cy=1,
        cx=1
    )
    res = lt_ctx.run(analysis)

    print(data)
    print("y", res["y"].raw_data)
    print("x", res["x"].raw_data)
    print("divergence", res["divergence"].raw_data)
    print("curl", res["curl"].raw_data)

    assert np.all(res["x"].raw_data == [[-1, -1, -1], [0, 0, 0], [1, 1, 1]])
    assert np.all(res["y"].raw_data == [[1, 0, -1], [1, 0, -1], [1, 0, -1]])

    assert np.all(res["divergence"].raw_data == 0)
    assert np.all(res["curl"].raw_data == -2)


def test_com_curl_flip(lt_ctx):
    data = np.zeros((3, 3, 3, 3), dtype=np.float32)
    for y in range(3):
        for x in range(3):
            data[y, x, 2-x, y] = 1
    data_flipped = np.flip(data, axis=2)
    dataset = MemoryDataSet(
        data=data_flipped,
        sig_dims=2,
    )
    analysis = lt_ctx.create_com_analysis(
        dataset=dataset,
        cy=1,
        cx=1,
        flip_y=True
    )
    res = lt_ctx.run(analysis)

    print("data", data)
    print("flipped", data_flipped)
    print("y", res["y"].raw_data)
    print("x", res["x"].raw_data)
    print("divergence", res["divergence"].raw_data)
    print("curl", res["curl"].raw_data)

    assert np.all(res["x"].raw_data == [[-1, -1, -1], [0, 0, 0], [1, 1, 1]])
    assert np.all(res["y"].raw_data == [[1, 0, -1], [1, 0, -1], [1, 0, -1]])

    assert np.all(res["divergence"].raw_data == 0)
    assert np.all(res["curl"].raw_data == -2)


def test_com_curl_rotate(lt_ctx):
    data = np.zeros((3, 3, 3, 3), dtype=np.float32)
    for y in range(3):
        for x in range(3):
            data[y, x, 2-x, y] = 1
    # Rotate 90 degrees clockwise
    data_90deg = np.zeros_like(data)
    for y in range(3):
        for x in range(3):
            data_90deg[:, :, x, 2-y] = data[:, :, y, x]
    dataset = MemoryDataSet(
        data=data_90deg,
        sig_dims=2,
    )
    analysis = lt_ctx.create_com_analysis(
        dataset=dataset,
        cy=1,
        cx=1,
        scan_rotation=-90.
    )
    res = lt_ctx.run(analysis)

    print("data", data)
    print("data_90deg", data_90deg)
    print("y", res["y"].raw_data)
    print("x", res["x"].raw_data)
    print("divergence", res["divergence"].raw_data)
    print("curl", res["curl"].raw_data)

    assert np.allclose(res["x"].raw_data, [[-1, -1, -1], [0, 0, 0], [1, 1, 1]])
    assert np.allclose(res["y"].raw_data, [[1, 0, -1], [1, 0, -1], [1, 0, -1]])

    assert np.allclose(res["divergence"].raw_data, 0)
    assert np.allclose(res["curl"].raw_data, -2)


def test_com_invalid_parameters(lt_ctx, ds_random):
    # mask_radius_inner depends on mask_radius being set, too,
    # to switch to annular mode:
    with pytest.raises(ValueError):
        lt_ctx.create_com_analysis(
            dataset=ds_random,
            cx=8,
            cy=8,
            mask_radius_inner=7,
        )


def test_com_valid_parameters(lt_ctx, ds_random):
    a = lt_ctx.create_com_analysis(
        dataset=ds_random,
        cx=8,
        cy=8,
        mask_radius=12,
        mask_radius_inner=7,
    )

    assert a.parameters['r'] == 12
    assert a.parameters['ri'] == 7


@pytest.mark.parametrize(
    'roi', (
        None,
        (slice(1, 4), slice(1, 4)),
        np.array([
            (True, True, True, True, False),
            (True, True, True, True, False),
            (True, True, True, True, False),
            (True, True, True, True, False),
            (False, False, False, False, False),
        ])
    )
)
def test_com_parameter_guess(lt_ctx, roi):
    data = np.zeros((5, 5, 5, 5), dtype=np.float32)
    # data with negative divergence and no curl
    for i in range(3):
        for j in range(3):
            data[i+1, j+1, 3-i, 3-j] = 1
    data[0, :, 2, 2] = 1
    data[4, :, 2, 2] = 1
    data[:, 0, 2, 2] = 1
    data[:, 4, 2, 2] = 1

    data_fliprot = data.transpose(0, 1, 3, 2)

    dataset = MemoryDataSet(
        data=data,
        sig_dims=2,
    )

    analysis = lt_ctx.create_com_analysis(
        dataset=dataset,
        cy=2,
        cx=2,
        scan_rotation=0.,
        flip_y=False
    )
    res = lt_ctx.run(analysis)

    print(res.divergence.raw_data)
    print(res.curl.raw_data)

    guess = guess_corrections(res.y.raw_data, res.x.raw_data, roi=roi)
    print(guess)

    g_rot, g_flip_y, g_cy, g_cx = guess
    # Check namedtuple
    assert guess.scan_rotation == g_rot
    assert guess.flip_y == g_flip_y
    assert guess.cy == g_cy
    assert guess.cx == g_cx

    assert g_rot == 0
    assert g_flip_y is False
    assert g_cy == 0
    assert g_cx == 0

    dataset_changed = MemoryDataSet(
        data=data_fliprot,
        sig_dims=2,
    )

    analysis_changed = lt_ctx.create_com_analysis(
        dataset=dataset_changed,
        cy=3,
        cx=1,
        scan_rotation=0.,
        flip_y=False,
    )
    res_changed = lt_ctx.run(analysis_changed)

    guess = guess_corrections(res_changed.y.raw_data, res_changed.x.raw_data, roi=roi)
    print(guess)

    g_rot, g_flip_y, g_cy, g_cx = guess

    # Transposing is equivalent to flipping and rotating 90°
    assert g_rot == 90
    assert g_flip_y is True
    assert g_cy == -1
    assert g_cx == 1

    # We apply the corrections
    analysis_corrected = lt_ctx.create_com_analysis(
        dataset=dataset_changed,
        cy=3+g_cy,
        cx=1+g_cx+1,
        scan_rotation=0.+g_rot,
        flip_y=(g_flip_y is not False),
    )
    res_corrected = lt_ctx.run(analysis_corrected)

    corrected_guess = guess_corrections(res_corrected.y.raw_data, res_corrected.x.raw_data, roi=roi)
    print(corrected_guess)
    print(res_corrected.divergence.raw_data)
    print(res_corrected.curl.raw_data)

    g_rot, g_flip_y, g_cy, g_cx = corrected_guess

    # Backtransform of the shift
    g_cy, g_cx = apply_correction(g_cy, g_cx, 90, True, forward=False)

    print(g_cy, g_cx)

    assert g_rot == 0
    assert g_flip_y is False
    assert np.allclose(g_cy, 0)
    assert np.allclose(g_cx, -1)


def test_apply_correction():
    y_centers = np.array((
        [1, 1],
        [-1, 0]
    ))
    x_centers = np.array((
        [0, 1],
        [1, 1]
    ))

    forward_1 = apply_correction(y_centers, x_centers, 90, False)
    forward_2 = apply_correction(y_centers, x_centers, 90, True)
    assert np.allclose(
        forward_1[0],
        np.array((
            [0, 1],
            [1, 1]
        ))
    )
    assert np.allclose(
        forward_1[1],
        np.array((
            [-1, -1],
            [1, 0]
        ))
    )
    assert np.allclose(
        forward_2[0],
        np.array((
            [0, 1],
            [1, 1]
        ))
    )
    assert np.allclose(
        forward_2[1],
        np.array((
            [1, 1],
            [-1, 0]
        ))
    )


@pytest.mark.parametrize(
    'angle', (0, -173, 23)
)
@pytest.mark.parametrize(
    'flip_y', (False, True)
)
def test_apply_correction_2(angle, flip_y):
    y_centers = np.array((
        [1, 2],
        [3, 4]
    ))
    x_centers = np.array((
        [8, 7],
        [6, 5]
    ))

    forward = apply_correction(y_centers, x_centers, angle, flip_y)
    back = apply_correction(forward[0], forward[1], angle, flip_y, forward=False)
    assert np.allclose(back[0], y_centers)
    assert np.allclose(back[1], x_centers)



from .base import (
    UDF, UDFMeta, UDFData, UDFFrameMixin, UDFTileMixin, UDFPartitionMixin,
    UDFPostprocessMixin, UDFPreprocessMixin, check_cast, UDFRunCancelled,
)
from .auto import AutoUDF


__all__ = [
    'UDF', 'UDFFrameMixin', 'UDFTileMixin', 'UDFPartitionMixin', 'UDFPostprocessMixin',
    'UDFPreprocessMixin', 'UDFMeta', 'UDFData', 'check_cast', 'AutoUDF', 'UDFRunCancelled',
]



import numpy as np

import libertem.udf.crystallinity as crystal
from libertem.analysis.apply_fft_mask import ApplyFFTMask

from libertem.io.dataset.memory import MemoryDataSet


def test_smoke(lt_ctx):
    """
    just check if the analysis runs without throwing exceptions:
    """
    data = np.zeros([3*3, 8, 8]).astype(np.float32)
    dataset = MemoryDataSet(data=data, tileshape=(1, 8, 8),
                            num_partitions=2, sig_dims=2)
    result = crystal.run_analysis_crystall(
        ctx=lt_ctx, dataset=dataset, rad_in=0, rad_out=1,
        real_center=(1, 1), real_rad=1)
    assert np.allclose(result['intensity'].data, np.zeros(data.shape[0]))


def test_simple_example(lt_ctx):
    # creating a dataset where 0:3 frames are strong crystalline, 3:6 frames are weak crystalline,
    #  6:9 frames are amourphous
    data = np.zeros([3*3, 5, 5]).astype(np.float32)
    # adding high intensity zero order peak for all frames
    data[:, 2, 2] = 7
    # adding strong non-zero order diffraction peaks for 0:3 frames
    data[0:3, 0, 0] = 2
    data[0:3, 4, 4] = 2
    # adding weak non-zero order diffraction peaks for 0:3 frames
    data[3:6, 2, 0] = 1
    data[3:6, 2, 4] = 1
    dataset = MemoryDataSet(data=data, tileshape=(1, 5, 5),
                            num_partitions=3, sig_dims=2)
    result = crystal.run_analysis_crystall(ctx=lt_ctx, dataset=dataset, rad_in=0, rad_out=3,
                                           real_center=(2, 2), real_rad=0)
    # check if values of integration in Fourier space after deleting of zero order diffraction peaks
    #  are zeros for amorphous frames
    assert np.allclose(result['intensity'].data[6:9], np.zeros([3]))
    # check if values of integration in Fourier space after deleting of zero order diffraction peaks
    #  are NOT zeros for strong crystalline frames
    assert (result['intensity'].data[0:3] > np.zeros([3])).all()
    # check if values of integration in Fourier space after deleting of zero order diffraction peaks
    #  are NOT zeros for weak crystalline frames
    assert (result['intensity'].data[3:6] > np.zeros([3])).all()
    # check if values of integration in Fourier space after deleting of zero order diffraction peaks
    #  are higher for strong crystalline frames than for weak crystalline frames
    assert (result['intensity'].data[0:3] > result['intensity'].data[3:6]).all()


def test_fft_mask_analysis(lt_ctx):
    # same setup as above
    data = np.zeros([3*3, 5, 5]).astype(np.float32)
    data[:, 2, 2] = 7
    data[0:3, 0, 0] = 2
    data[0:3, 4, 4] = 2
    data[3:6, 2, 0] = 1
    data[3:6, 2, 4] = 1
    dataset = MemoryDataSet(data=data, tileshape=(1, 5, 5),
                            num_partitions=3, sig_dims=2)
    analysis = ApplyFFTMask(dataset=dataset, parameters={
        'rad_in': 0,
        'rad_out': 3,
        'real_centerx': 2,
        'real_centery': 2,
        'real_rad': 0,
    })
    result = lt_ctx.run(analysis)

    assert np.allclose(result['intensity'].raw_data[6:9], np.zeros([3]))
    assert (result['intensity'].raw_data[0:3] > np.zeros([3])).all()
    assert (result['intensity'].raw_data[3:6] > np.zeros([3])).all()
    assert (result['intensity'].raw_data[0:3] > result['intensity'].raw_data[3:6]).all()



import logging
import warnings
from contextlib import contextmanager

import numpy as np
import matplotlib.pyplot as plt
from jupyter_ui_poll import ui_events

from .base import Live2DPlot


logger = logging.getLogger(__name__)

try:
    with ui_events() as poll:
        pass
except Exception:
    logger.info(
            'Deactivating Jupyter UI event polling, '
            'possibly not running in IPython?',
            exc_info=True
        )

    # Replacing jupyter_ui_poll with a dummy
    @contextmanager
    def ui_events():
        yield lambda x: None


class MPLLive2DPlot(Live2DPlot):
    """
    Matplotlib-based live plot

    .. versionadded:: 0.7.0

    Parameters
    ----------
    dataset : DataSet
        The dataset on which the UDF will be run. This allows
        to determine the shape of the plots for initialization.
    udf : UDF
        The UDF instance this plot is associated to. This needs to be
        the same instance that is passed to
        :meth:`~libertem.api.Context.run_udf`.
    roi : numpy.ndarray or None
        Region of interest (ROI) that the UDF will
        be run on. This is necessary for UDFs where the `extra_shape`
        parameter of result buffers is a function of the ROI, such as
        :class:`~libertem.udf.raw.PickUDF`.
    channel : misc
        Indicate the channel to be plotted.

        - :code:`None`: The first plottable (2D) channel of the UDF is plotted.
        - :code:`str`: The UDF result buffer name that should be plotted.
        - :code:`tuple(str, function(ndarray) -> ndarray)`: The UDF result buffer name that
          should be plotted together with a function that extracts a plottable result
        - :code:`function(udf_result, damage) -> (ndarray, damage)`: Function that derives a
          plottable 2D ndarray and damage indicator from the full
          UDF results and the processed nav space. See :ref:`plotting` for more details!
    title : str
        The plot title. By default UDF class name and channel name.
    min_delta : float
        Minimum time span in seconds between updates to reduce overheads for slow plotting.
    udfresult : UDFResults, optional
        UDF result to initialize the plot data and determine plot shape. If None (default),
        this is determined using :meth:`~libertem.udf.base.UDFRunner.dry_run` on the dataset,
        UDF and ROI. This parameter allows re-using buffers to avoid unnecessary dry runs.

    **kwargs
        Passed on to :code:`imshow`
    """
    def __init__(
            self, dataset, udf, roi=None, channel=None, title=None, min_delta=0.5, udfresult=None,
            **kwargs,
    ):
        super().__init__(
            dataset=dataset,
            udf=udf,
            roi=roi,
            channel=channel,
            title=title,
            min_delta=min_delta,
            udfresult=udfresult
        )
        self.kwargs = kwargs
        self.fig = None
        self.axes = None
        self.im_obj = None

    def display(self):
        with ui_events() as poll:
            self.fig, self.axes = plt.subplots()
            self.im_obj = self.axes.imshow(self.data, **self.kwargs)
            # Set values compatible with log norm
            self.im_obj.norm.vmin = 1
            self.im_obj.norm.vmax = 1 + 1e-12
            self.axes.set_title(self.title)
            self.fig.show()
            poll(1000)

    def update(self, damage, force=False):
        """
        Update the plot based on `self.data`.

        Parameters
        ----------
        force : bool
            Force an update, disabling any throttling mechanisms
        """
        # Nothing to draw, not displayed
        if self.im_obj is None:
            assert self.fig is None
            assert self.axes is None
            warnings.warn(
                "Plot is not displayed, not plotting. "
                "Call display() to display the plot."
            )
            return

        i_o = self.im_obj
        i_o.set_data(self.data)
        valid_data = self.data[damage]
        valid_data = valid_data[np.isfinite(valid_data)]
        if len(valid_data) > 0:
            i_o.norm.vmin = np.min(valid_data)
            i_o.norm.vmax = np.max(valid_data)
        with ui_events() as poll:
            i_o.changed()
            self.fig.canvas.draw_idle()
            self.fig.canvas.flush_events()
            poll(1000)



import os
import typing
import logging
import warnings

from ncempy.io.dm import fileDM
import numpy as np

from libertem.common.math import prod
from libertem.common import Shape
from libertem.io.dataset.base.file import OffsetsSizes
from libertem.common.messageconverter import MessageConverter
from .base import (
    DataSet, FileSet, BasePartition, DataSetException, DataSetMeta, File,
    IOBackend,
)

log = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from numpy import typing as nt


class SingleDMDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/DMDatasetParams.schema.json",
        "title": "DMDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "DM"},
            "path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
            "io_backend": {
                "enum": IOBackend.get_supported(),
            },
            "force_c_order": {"type": "boolean"},
        },
        "required": ["type", "path"]
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        for k in ["nav_shape", "sig_shape", "sync_offset", "force_c_order"]:
            if k in raw_data:
                data[k] = raw_data[k]
        return data


class StackedDMDatasetParams(MessageConverter):
    SCHEMA: dict = {}

    def convert_from_python(self, raw_data):
        return super().convert_from_python(raw_data)

    def convert_to_python(self, raw_data):
        return super().convert_to_python(raw_data)


def _get_metadata(path):
    fh = fileDM(path, on_memory=True)
    if fh.numObjects == 1:
        idx = 0
    else:
        idx = 1
    return {
        'offset': fh.dataOffset[idx],
        'zsize': fh.zSize[idx],
    }


class StackedDMFile(File):
    def get_array_from_memview(self, mem: memoryview, slicing: OffsetsSizes):
        mem = mem[slicing.file_offset:-slicing.skip_end]
        res = np.frombuffer(mem, dtype="uint8")
        itemsize = np.dtype(self._native_dtype).itemsize
        sigsize = int(prod(self._sig_shape))
        cutoff = 0
        cutoff += (
            self.num_frames * itemsize * sigsize
        )
        res = res[:cutoff]
        return res.view(dtype=self._native_dtype).reshape(
            (self.num_frames, -1)
        )[:, slicing.frame_offset:slicing.frame_offset + slicing.frame_size]


class DMFileSet(FileSet):
    pass


class DMDataSet(DataSet):
    """
    Factory class for DigitalMicrograph file datasets

     - Passing either the :code:`files` kwarg or a tuple/list as first
       argument will create an instance of :class:`StackedDMDataSet`
     - Passing either the :code:`path` kwarg or any other object
       as first argument will create a :class:`SingleDMDataSet`

    This class is necessary to handle the difference in signatures and
    behaviours of the two DM dataset implementations, but these may
    later be fused if a markup format for multi-file datasets is implemented

    This class implements the methods necessary to expose a DMDataSet in
    the web GUI, which it does by deferring to SingleDMDataSet. At this
    time multi-file datasets are not supported in the UI.

    NOTE this way of generating the subclasses breaks deeper
    subclassing, as this __new__ method will always instantiate
    a SingleDMDataSet or StackedDMDataSet, and not a subclass of
    either of these. This could potentially be improved by using
    the .__instance_subclass__() staticmethod to register the
    subclasses and what they inherit from.
    """
    def __new__(cls, *args, **kwargs):
        # delayed here to avoid circular reference
        from .dm_single import SingleDMDataSet
        if 'path' in kwargs:
            subclass = SingleDMDataSet
        elif 'files' in kwargs:
            subclass = StackedDMDataSet
        elif args and isinstance(args[0], (list, tuple)):
            subclass = StackedDMDataSet
        else:
            subclass = SingleDMDataSet
        return super().__new__(subclass)

    @classmethod
    def get_supported_extensions(cls):
        return {"dm3", "dm4"}

    @classmethod
    def get_msg_converter(cls):
        return SingleDMDatasetParams

    @classmethod
    def detect_params(cls, path, executor):
        # delayed here to avoid circular reference
        from .dm_single import SingleDMDataSet
        return SingleDMDataSet.detect_params(path, executor)


class StackedDMDataSet(DMDataSet):
    """
    Reader for stacks of DM3/DM4 files.

    Note
    ----
    This DataSet is not supported in the GUI yet, as the file dialog needs to be
    updated to `properly handle opening series
    <https://github.com/LiberTEM/LiberTEM/issues/498>`_.

    Note
    ----
    Single-file 3/4D DM datasets are supported through the
    :class:`~libertem.io.datasets.dm_single.SingleDMDataSet` class.

    Note
    ----
    You can use the PyPI package `natsort <https://pypi.org/project/natsort/>`_
    to sort the filenames by their numerical components, this is especially useful
    for filenames without leading zeros.

    Parameters
    ----------

    files : List[str]
        List of paths to the files that should be loaded. The order is important,
        as it determines the order in the navigation axis.

    nav_shape : Tuple[int, ...] or None
        By default, the files are loaded as a 3D stack. You can change this
        by specifying the nav_shape, which reshapes the navigation dimensions.
        Raises a `DataSetException` if the shape is incompatible with the data
        that is loaded.

    sig_shape: Tuple[int, ...], optional
        Signal/detector size (height, width)

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start

    same_offset : bool
        When reading a stack of dm3/dm4 files, it can be expensive to read in
        all the metadata from all files, which we currently only use for
        getting the offsets and sizes of the main data in each file. If you
        absolutely know that the offsets and sizes are the same for all files,
        you can set this parameter and we will skip reading all metadata but
        the one from the first file.
    """
    def __init__(self, files=None, scan_size=None, same_offset=False, nav_shape=None,
                 sig_shape=None, sync_offset=0, io_backend=None):
        super().__init__(io_backend=io_backend)
        self._meta = None
        self._same_offset = same_offset
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset
        # handle backwards-compatability:
        if scan_size is not None:
            warnings.warn(
                "scan_size argument is deprecated. please specify nav_shape instead",
                FutureWarning
            )
            if nav_shape is not None:
                raise ValueError("cannot specify both scan_size and nav_shape")
            self._nav_shape = tuple(scan_size)
        self._filesize = None
        self._files = files
        if not isinstance(files, (list, tuple)):
            raise DataSetException("files argument must be an iterable\
                                    of file paths, recieved {type(files)}")
        if len(files) == 0:
            raise DataSetException("need at least one file as input!")
        self._fileset = None
        # per-file cached attributes:
        self._z_sizes = {}
        self._offsets = {}

    def __new__(cls, *args, **kwargs):
        '''
        Skip the superclasse's :code:`__new__()` method.

        Instead, go straight to the grandparent. That disables the
        :class:`DMDataSet` type determination magic. Otherwise unpickling will
        always yield a :class:`SingleDMDataSet` since this class inherits the
        parent's :code:`__new__()` method and unpickling calls it without
        parameters, making it select :class:`SingleDMDataSet`.

        It mimics calling the superclass :code:`__new__(cls)` without additional
        parameters, just like the parent's method.
        '''
        return DataSet.__new__(cls)

    def _get_sig_shape_and_native_dtype(self):
        first_fn = self._get_files()[0]
        first_file = fileDM(first_fn, on_memory=True)
        if first_file.numObjects == 1:
            idx = 0
        else:
            idx = 1
        try:
            raw_dtype = first_file._DM2NPDataType(first_file.dataType[idx])
            native_sig_shape = (first_file.ySize[idx], first_file.xSize[idx])
        except IndexError as e:
            raise DataSetException("could not determine dtype or signal shape") from e
        return native_sig_shape, raw_dtype

    def _get_fileset(self):
        start_idx = 0
        files = []
        for fn in self._get_files():
            z_size = self._z_sizes[fn]
            f = StackedDMFile(
                path=fn,
                start_idx=start_idx,
                end_idx=start_idx + z_size,
                sig_shape=self._meta.shape.sig,
                native_dtype=self._meta.raw_dtype,
                file_header=self._offsets[fn],
            )
            files.append(f)
            start_idx += z_size
        return DMFileSet(files)

    def _get_files(self):
        return self._files

    def _get_filesize(self):
        return sum(
            os.stat(p).st_size
            for p in self._get_files()
        )

    def initialize(self, executor):
        self._filesize = executor.run_function(self._get_filesize)
        if self._same_offset:
            metadata = executor.run_function(_get_metadata, self._get_files()[0])
            self._offsets = {
                fn: metadata['offset']
                for fn in self._get_files()
            }
            self._z_sizes = {
                fn: metadata['zsize']
                for fn in self._get_files()
            }
        else:
            metadata = dict(zip(
                self._get_files(),
                executor.map(_get_metadata, self._get_files()),
            ))
            self._offsets = {
                fn: metadata[fn]['offset']
                for fn in self._get_files()
            }
            self._z_sizes = {
                fn: metadata[fn]['zsize']
                for fn in self._get_files()
            }
        self._image_count = sum(self._z_sizes.values())
        if self._nav_shape is None:
            self._nav_shape = (sum(self._z_sizes.values()),)
        native_sig_shape, native_dtype = executor.run_function(self._get_sig_shape_and_native_dtype)
        if self._sig_shape is None:
            self._sig_shape = tuple(native_sig_shape)
        elif int(prod(self._sig_shape)) != int(prod(native_sig_shape)):
            raise DataSetException(
                "sig_shape must be of size: %s" % int(prod(native_sig_shape))
            )
        shape = self._nav_shape + self._sig_shape
        self._nav_shape_product = int(prod(self._nav_shape))
        self._sync_offset_info = self.get_sync_offset_info()
        self._meta = DataSetMeta(
            shape=Shape(shape, sig_dims=len(self._sig_shape)),
            raw_dtype=native_dtype,
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        self._fileset = executor.run_function(self._get_fileset)
        return self

    @classmethod
    def get_supported_extensions(cls):
        return {"dm3", "dm4"}

    @classmethod
    def get_msg_converter(cls) -> type[MessageConverter]:
        return StackedDMDatasetParams

    @classmethod
    def detect_params(cls, path, executor):
        # FIXME: this doesn't really make sense for file series
        # pl = path.lower()
        # if pl.endswith(".dm3") or pl.endswith(".dm4"):
        #     return {
        #         "parameters": {
        #             "files": [path]
        #         },
        #     }
        return False

    @property
    def dtype(self) -> "nt.DTypeLike":
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    def check_valid(self):
        first_fn = self._get_files()[0]
        try:
            with fileDM(first_fn, on_memory=True):
                pass
            return True
        except OSError as e:
            raise DataSetException("invalid dataset: %s" % e)

    def get_partitions(self):
        for part_slice, start, stop in self.get_slices():
            yield BasePartition(
                meta=self._meta,
                partition_slice=part_slice,
                fileset=self._fileset,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )

    def __repr__(self):
        return "<DMDataSet for a stack of %d files>" % (len(self._get_files()),)



import random

import numpy as np
import pytest

from libertem.udf.sum import SumUDF
from sparseconverter import (
    CUPY_SCIPY_CSR, NUMPY, SCIPY_COO, SPARSE_COO, SPARSE_GCXS, get_backend, for_backend
)

from utils import _mk_random, set_device_class


# SumUDF supports sparse input and CuPy
# Test that densification for CuPy and sparse input on NumPy works
class SparseInputSumUDF(SumUDF):
    def get_backends(self):
        return (SPARSE_COO, SPARSE_GCXS, SCIPY_COO, CUPY_SCIPY_CSR)

    def process_tile(self, tile):
        format = get_backend(tile)
        if self.meta.device_class == 'cpu':
            assert format in (SPARSE_COO, SPARSE_GCXS, SCIPY_COO)
        else:
            assert format in (CUPY_SCIPY_CSR, )
        return super().process_tile(tile)


# Make sure we actually use CuPy
class CupyInputSumUDF(SparseInputSumUDF):
    def process_tile(self, tile):
        assert self.meta.device_class == 'cuda'
        return super().process_tile(tile)


# Test densification due to UDF format requirements
class DenseInputSumUDF(SumUDF):
    def get_backends(self):
        return (self.BACKEND_NUMPY, self.BACKEND_CUPY)

    def process_tile(self, tile):
        backend = get_backend(tile)
        assert backend in (self.BACKEND_NUMPY, self.BACKEND_CUPY)
        return super().process_tile(tile)


# Test sparsification due to UDF format requirements
class OnlySparseSumUDF(SumUDF):
    def get_backends(self):
        return self.BACKEND_SPARSE_GCXS

    def process_tile(self, tile):
        backend = get_backend(tile)
        assert backend == self.BACKEND_SPARSE_GCXS
        return super().process_tile(tile)


@pytest.mark.parametrize(
    'format', (NUMPY, SPARSE_COO, SPARSE_GCXS)
)
def test_sparse(lt_ctx, format):
    '''
    Test that constraints are observed and selection logic works
    with different sets of UDFs.
    '''
    data = _mk_random((7, 11, 13), array_backend=format)
    ds = lt_ctx.load('memory', data=data)
    udfs = [OnlySparseSumUDF(), DenseInputSumUDF(), SumUDF(), SumUDF()]
    if format != NUMPY:
        # Make sure sparse input is used if the dataset is sparse
        udfs.append(SparseInputSumUDF())
    for i in range(5):
        subset = random.sample(udfs, k=random.randint(1, len(udfs)))
        random.shuffle(subset)
        res = lt_ctx.run_udf(
            dataset=ds,
            udf=subset
        )
        ref = for_backend(data.sum(axis=0), NUMPY)
        for r in res:
            assert np.allclose(ref, r['intensity'].raw_data)


def test_on_cuda(lt_ctx):
    with set_device_class('cupy'):
        data = _mk_random((7, 11, 13), array_backend=SPARSE_COO)
        ds = lt_ctx.load('memory', data=data)
        # Densifies input on CuPy
        res1 = lt_ctx.run_udf(dataset=ds, udf=[CupyInputSumUDF()])
        res2 = lt_ctx.run_udf(dataset=ds, udf=[OnlySparseSumUDF()])
        ref = data.sum(axis=0).todense()
        assert np.allclose(ref, res1[0]['intensity'].raw_data)
        assert np.allclose(ref, res2[0]['intensity'].raw_data)



import logging
from typing import TYPE_CHECKING

import tornado.web
from opentelemetry import trace

from libertem.web.engine import JobEngine
from libertem.executor.base import AsyncAdapter
from .messages import Message
from .base import log_message
from .state import SharedState

log = logging.getLogger(__name__)
tracer = trace.get_tracer(__name__)

if TYPE_CHECKING:
    from libertem.web.events import EventRegistry


class ConnectHandler(tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry: "EventRegistry"):
        self.state = state
        self.event_registry = event_registry
        self.engine = JobEngine(state, event_registry)

    async def get(self):
        log.info("ConnectHandler.get")
        try:
            await self.state.executor_state.get_executor()
            params = self.state.executor_state.get_cluster_params()
            # TODO: extract into Message class
            self.write({
                "status": "ok",
                "connection": params["connection"],
            })
        except RuntimeError:  # TODO: exception class is too generic
            # TODO: extract into Message class
            self.write({
                "status": "disconnected",
                "connection": {},
            })

    async def put(self):
        with tracer.start_as_current_span("ConnectHandler.put"):
            await self._do_connect()

    async def _do_connect(self):
        # TODO: extract json request data stuff into mixin?
        request_data = tornado.escape.json_decode(self.request.body)
        pool = AsyncAdapter.make_pool()
        with tracer.start_as_current_span("executor setup"):
            try:
                executor = await self.state.executor_state.make_executor(request_data, pool)
            except Exception as e:
                msg = Message().cluster_conn_error(msg=str(e))
                log_message(msg, exception=True)
                self.set_status(500)
                self.write(msg)
                return None
            await self.state.executor_state.set_executor(executor, request_data)
        await self.state.dataset_state.verify()
        datasets = await self.state.dataset_state.serialize_all()
        msg = Message().initial_state(
            jobs=self.state.job_state.serialize_all(),
            datasets=datasets, analyses=self.state.analysis_state.serialize_all(),
            compound_analyses=self.state.compound_analysis_state.serialize_all(),
        )
        log_message(msg)
        # FIXME: don't broadcast, only send to the websocket that matches this HTTP connection
        # (is this even possible?)
        self.event_registry.broadcast_event(msg)
        await self.engine.send_existing_job_results()
        self.write({
            "status": "ok",
            "connection": request_data['connection'],
        })



import os

from libertem.io.utils import get_owner_name


def test_get_owner_name():
    owner = get_owner_name(__file__, os.stat(__file__))
    assert isinstance(owner, str)
    assert len(owner) > 0



import os
import json
from collections import defaultdict

import numpy as np

from libertem.common.math import prod
from .base import (
    DataSet, BasePartition, DataSetMeta, DataSetException,
    WritablePartition, WritableDataSet
)
from .raw import RawFile, RawFileSet
from libertem.common.scheduler import WorkerSet
from libertem.io.writers.base import WriteHandle
from libertem.common import Shape, Slice


class ClusterDataSet(WritableDataSet, DataSet):
    """
    ClusterDataSet: a distributed RAW data set

     * to be used for the cache, for live acquisition, and for simulation integration
     * each node has a directory for a ClusterDataSet
     * the directory contains partitions, each its own raw file
     * information about the structure is saved as a json sidecar file
     * A ClusterDataSet dataset can be incomplete, that is, it can miss complete partitions
       (but partitions themselves are guaranteed to be complete once they have their final filename)
     * use cases for incomplete datasets:
        * each node only caches the partitions it is responsible for
        * partial acquisitions support
     * missing partitions can later be written
     * file names and structure/partitioning are deterministic
     * assumption: all workers on a single host share the dataset

    Parameters
    ----------

    path : str
        Absolute filesystem base path, pointing to an existing directory.
        Assumes a uniform setup (same absolute path used on all nodes)

    structure : PartitionStructure
        Partitioning structure instance. Must be specified when creating a new dataset.
    """
    def __init__(self, path, structure=None, io_backend=None):
        super().__init__(io_backend=io_backend)
        self._path = path
        self._dtype = structure.dtype
        self._structure = structure
        self._meta = DataSetMeta(
            shape=structure.shape,
            raw_dtype=np.dtype(structure.dtype),
            sync_offset=0,
            image_count=int(prod(structure.shape.nav)),
        )
        self._executor = None

    def initialize(self, executor):
        """
        Initialize is running on the master node, but we have
        access to the executor.
        """
        # 1) create sidecar files on hosts that don't have them yet
        executor.run_each_host(self._ensure_sidecar)
        # 2) check that all sidecar files that exist are equal
        sidecars = executor.run_each_host(self._read_sidecar)
        given_structure = self._structure.serialize()
        if not all(s == given_structure for s in sidecars.values()):
            print(sidecars.values())
            print(given_structure)
            raise DataSetException(
                "inconsistent sidecars, please inspect {} on each node".format(
                    self._sidecar_path(),
                )
            )
        self._executor = executor
        return self

    def _get_fileset(self):
        return RawFileSet([
            RawFile(
                path=self._get_path_for_idx(idx),
                start_idx=start_idx,
                end_idx=end_idx,
                sig_shape=self.shape.sig,
                native_dtype=self._meta.raw_dtype,
            )
            for (idx, (start_idx, end_idx)) in enumerate(self._structure.slices)
        ])

    def _get_path_for_idx(self, idx):
        # FIXME: possibly different paths per node
        return os.path.join(self._path, "parts", "partition-%08d" % idx)

    @property
    def dtype(self):
        return self._dtype

    @property
    def shape(self):
        return self._meta.shape

    def _ensure_sidecar(self):
        """
        run on each node on initialization
        """
        os.makedirs(os.path.join(self._path, "parts"), exist_ok=True)
        if not os.path.exists(self._sidecar_path()):
            self._write_sidecar()

    def _sidecar_path(self):
        return os.path.join(self._path, "structure.json")

    def _read_sidecar(self):
        """
        run on each node on initialization
        """
        with open(self._sidecar_path()) as fh:
            return json.load(fh)

    def _write_sidecar(self):
        """
        run on each node on initialization, if sidecar doesn't exist
        """
        with open(self._sidecar_path(), "w") as fh:
            json.dump(self._structure.serialize(), fh)

    def check_valid(self):
        if not os.path.exists(self._path) or not os.path.isdir(self._path):
            raise DataSetException("path %s does not exist or is not a directory" % self._path)

    @classmethod
    def detect_params(cls, path, executor):
        # TODO: read sidecar file etc.?
        return False

    @classmethod
    def get_msg_converter(cls):
        raise NotImplementedError()

    def get_diagnostics(self):
        return []

    def get_partitions(self):
        idx_to_workers = self._get_all_workers()
        fileset = self._get_fileset()
        for (idx, (start_idx, end_idx)) in enumerate(self._structure.slices):
            part_slice = Slice(
                origin=(start_idx,) + tuple([0] * self.shape.sig.dims),
                shape=Shape(((end_idx - start_idx),) + tuple(self.shape.sig),
                            sig_dims=self.shape.sig.dims)
            )
            yield ClusterDSPartition(
                path=self._get_path_for_idx(idx),
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start_idx,
                num_frames=end_idx - start_idx,
                workers=idx_to_workers[idx] or None,
                io_backend=self.get_io_backend(),
            )

    def _get_all_workers(self):
        """
        returns a mapping idx -> workers
        """
        if self._executor is None:
            raise RuntimeError(
                "invalid state: _get_all_workers needs access to the executor"
            )

        paths = [
            (idx, self._get_path_for_idx(idx))
            for (idx, slice_) in enumerate(self._structure.slices)
        ]

        def _check_paths():
            return [
                idx
                for idx, path in paths
                if os.path.exists(path)
            ]

        host_to_idxs = self._executor.run_each_host(_check_paths)
        workers = self._executor.get_available_workers()

        # invert: idx -> workers
        result = defaultdict(lambda: WorkerSet([]))
        for host, idxs in host_to_idxs.items():
            for idx in idxs:
                result[idx] = result[idx].extend(workers.get_by_host(host))
        return result


class ClusterDSPartition(WritablePartition, BasePartition):
    def __init__(self, path, workers, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._path = path
        self._workers = workers

    def get_write_handle(self):
        # FIXME: make example executable? involves temporary directory, source dataset, ...
        # TODO: Direct I/O writing support
        """
        Get a handle to write to this partition. Current rules:

        1) You can only write a complete partition at once, which is then immutable afterwards.
           If you want to change something, you need to write the whole partition again.
        2) Once the `with`-block is exited successfully, the data is written down to disk. If
           there is an error while writing the data, the partition will not be moved into its final
           place and it will be missing from the data set. There cannot be a "parially written"
           partition.

        Example
        -------
        >>> with dest_part.get_write_handle() as wh:  #  doctest: +SKIP
        ...     for tile in wh.write_tiles(source_part.get_tiles()):
        ...         pass  # do something with `tile`
        """
        tmp_path = os.path.dirname(self._path)
        return WriteHandle(
            path=self._path,
            tmp_base_path=tmp_path,
            dtype=self.dtype,
            part_slice=self.slice,
        )

    def delete(self):
        """
        delete this partition. needs to run on all nodes that have this partition
        """
        if os.path.exists(self._path):
            os.unlink(self._path)

    def _have_data(self):
        return os.path.exists(self._path)

    def get_locations(self):
        return self._workers

    def get_canonical_path(self):
        return os.path.realpath(self._path)



import pytest
import websockets


paths = [
    '/',
    "/api/datasets/detect/",
    "/api/datasets/schema/",
]


@pytest.mark.parametrize(
    'path', paths,
)
@pytest.mark.parametrize(
    'method', [
        'GET',
        'POST',
        'PUT',
        'DELETE',
    ]
)
@pytest.mark.asyncio
async def test_missing_token(base_url, path, method, http_client):
    full_url_without_token = base_url + path

    async with http_client.request(method, full_url_without_token) as resp:
        assert resp.status == 400 or resp.status == 405, \
            f"response code must be 'bad request' or 'method not allowed', got '{resp.status}'"


@pytest.mark.parametrize(
    'path', paths,
)
@pytest.mark.parametrize(
    'method', [
        'GET',
        'POST',
        'PUT',
        'DELETE',
    ]
)
@pytest.mark.asyncio
async def test_wrong_token(base_url, path, method, http_client):
    full_url_wrong_token = base_url + path + "?token=wrong"

    async with http_client.request(method, full_url_wrong_token) as resp:
        assert resp.status == 400 or resp.status == 405, \
            f"response code must be 'bad request' or 'method not allowed', got '{resp.status}'"


@pytest.mark.asyncio
async def test_token_for_ws_events(base_url, http_client, server_port):
    # fails without token:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/"
    with pytest.raises(websockets.InvalidStatusCode):
        async with websockets.connect(ws_url):
            pass


@pytest.mark.asyncio
async def test_no_token_server(base_url_no_token, http_client):
    # works without token:
    full_url = base_url_no_token + "/"
    async with http_client.get(full_url) as resp:
        assert resp.status == 200


# slow variants of some of the tests above:

paths_full = paths + [
    "/api/datasets/ds-uuid-here/",
    "/api/browse/localfs/",
    "/api/jobs/job-id-here/",
    "/api/compoundAnalyses/compound-uuid-here/analyses/analysis-uuid-here/",
    "/api/compoundAnalyses/compiund-uuid-here/analyses/analysis-uuid-here/download/HDF5/",
    "/api/compoundAnalyses/uuid-here/copy/notebook/",
    "/api/compoundAnalyses/uuid-here/download/notebook/",
    "/api/compoundAnalyses/uuid-here/",
    "/api/events/",
    "/api/shutdown/",
    "/api/config/",
    "/api/config/cluster/",
    "/api/config/connection/",
]


@pytest.mark.web_api
@pytest.mark.parametrize(
    'path', paths_full,
)
@pytest.mark.parametrize(
    'method', [
        'GET',
        'POST',
        'PUT',
        'DELETE',
        'HEAD',
        'PATCH',
        'OPTIONS',
    ]
)
@pytest.mark.asyncio
async def test_wrong_token_full(base_url, path, method, http_client):
    full_url_wrong_token = base_url + path + "?token=wrong"

    async with http_client.request(method, full_url_wrong_token) as resp:
        assert resp.status == 400 or resp.status == 405, \
            f"response code must be 'bad request' or 'method not allowed', got '{resp.status}'"


@pytest.mark.web_api
@pytest.mark.parametrize(
    'path', paths_full,
)
@pytest.mark.parametrize(
    'method', [
        'GET',
        'POST',
        'PUT',
        'DELETE',
        'HEAD',
        'PATCH',
        'OPTIONS',
    ]
)
@pytest.mark.asyncio
async def test_missing_token_full(base_url, path, method, http_client):
    full_url_without_token = base_url + path

    async with http_client.request(method, full_url_without_token) as resp:
        assert resp.status == 400 or resp.status == 405, \
            f"response code must be 'bad request' or 'method not allowed', got '{resp.status}'"



import pytest

from utils import assert_msg
from aio_utils import create_connection

pytestmark = [pytest.mark.web_api]


@pytest.mark.asyncio
async def test_shutdown(base_url, http_client, server_port, default_token):
    await create_connection(base_url, http_client, token=default_token)

    print("checkpoint 1")

    url = f"ws://127.0.0.1:{server_port}/api/shutdown/?token={default_token}"
    async with http_client.delete(url) as resp:
        print("checkpoint 2")
        assert resp.status == 200
        assert_msg(await resp.json(), "SERVER_SHUTDOWN")



from typing import Optional
from libertem.common import Shape
from libertem.common.math import prod

try:
    import pwd

    def get_owner_name(full_path, stat):
        try:
            name = pwd.getpwuid(stat.st_uid).pw_name
        except KeyError:
            name = stat.st_uid
        return name
# Assume that we are on Windows
# TODO do a proper abstraction layer if this simple solution doesn't work anymore
except ModuleNotFoundError:
    from libertem.common.win_tweaks import get_owner_name  # noqa: F401


def get_partition_shape(
    dataset_shape: Shape,
    target_size_items: int,
    min_num: Optional[int] = None
) -> tuple[int, ...]:
    """
    Calculate partition shape for the given ``target_size_items``

    Parameters
    ----------

    dataset_shape
        "native" dataset shape

    target_size_items
        target partition size in number of items/pixels

    min_num
        minimum number of partitions
    """
    sig_size = dataset_shape.sig.size
    current_p_shape: tuple[int, ...] = ()

    if min_num is None:
        min_num = 1

    target_size_items = min(target_size_items, int(dataset_shape.size // min_num))

    for dim in reversed(tuple(dataset_shape.nav)):
        proposed_shape = (dim,) + current_p_shape
        proposed_size = prod(proposed_shape) * sig_size
        if proposed_size <= target_size_items:
            current_p_shape = proposed_shape
        else:
            overshoot = proposed_size / target_size_items
            last_size = max(1, int(dim // overshoot))
            current_p_shape = (last_size,) + current_p_shape
            break

    res = tuple([1] * (len(dataset_shape.nav) - len(current_p_shape))) + current_p_shape
    return res



import os
import copy
import typing
import itertools
import logging
import asyncio
import time
import contextlib

import psutil

import libertem
from libertem.api import Context
from libertem.analysis.base import AnalysisResultSet
from libertem.common.executor import JobExecutor
from libertem.common.async_utils import sync_to_async
from libertem.executor.base import AsyncAdapter
from libertem.executor.dask import DaskJobExecutor
from libertem.io.dataset.base import DataSetException, DataSet
from libertem.io.writers.results.base import ResultFormatRegistry
from libertem.io.writers.results import formats  # NOQA
from libertem.udf.base import UDFResults
from libertem.utils import devices
from .models import (
    AnalysisDetails, AnalysisInfo, AnalysisResultInfo, CompoundAnalysisInfo, AnalysisParameters,
    DatasetInfo, JobInfo, SerializedJobInfo,
)
from .helpers import create_executor
from .event_bus import EventBus

log = logging.getLogger(__name__)


class ExecutorState:
    """
    Executor management for the web API. This class is used by `SharedState` to
    manage executor instances and related state.  Executors passed into
    `set_executors` will be taken over, and managed directly by this class.

    Be sure to explicitly call the `shutdown` method to clean up, otherwise
    there may be dangling resources preventing a clean shutdown.
    """
    def __init__(
        self,
        event_bus: EventBus,
        loop: typing.Optional[asyncio.AbstractEventLoop] = None,
        snooze_timeout: typing.Optional[float] = None,
    ):
        self.executor = None
        self.cluster_params = {}
        self.cluster_details: typing.Optional[list] = None
        self.context: typing.Optional[Context] = None
        self._event_bus = event_bus
        self._snooze_lock = asyncio.Lock()
        self._snooze_timeout = snooze_timeout
        self._pool = AsyncAdapter.make_pool()
        self._is_snoozing = False
        self._last_activity = time.monotonic()
        self._snooze_check_interval = min(
            30.0,
            snooze_timeout and (snooze_timeout * 0.1) or 30.0,
        )
        self._snooze_task = asyncio.ensure_future(self._snooze_check_task())
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop = loop
        self._keep_alive = 0
        self.local_directory = "dask-worker-space"
        self.preload: tuple[str, ...] = ()

    def set_preload(self, preload: tuple[str, ...]) -> None:
        self.preload = preload

    def get_preload(self) -> tuple[str, ...]:
        return self.preload

    def set_local_directory(self, local_directory: str) -> None:
        if local_directory is not None:
            self.local_directory = local_directory

    def get_local_directory(self):
        return self.local_directory

    def _update_last_activity(self):
        self._last_activity = time.monotonic()
        log.debug("_update_last_activity")

    @contextlib.contextmanager
    def keep_alive(self):
        self._update_last_activity()
        self._keep_alive += 1
        try:
            yield
        finally:
            self._keep_alive -= 1
            self._update_last_activity()

    async def snooze(self):
        async with self._snooze_lock:
            if self.executor is None:
                log.debug("not snoozing: no executor")
                return
            if self._keep_alive > 0:
                log.debug("not snoozing: _keep_alive=%d", self._keep_alive)
                return
            log.info("Snoozing...")
            from .messages import Message
            msg = Message().snooze("snoozing")
            self._event_bus.send(msg)
            self.executor.ensure_sync().close()
            self.context = None
            self.executor = None
            self._is_snoozing = True

    async def unsnooze(self):
        async with self._snooze_lock:
            if not self._is_snoozing:
                return
            log.info("Unsnoozing...")
            from .messages import Message
            msg = Message().unsnooze("unsnoozing")
            self._event_bus.send(msg)
            executor = await self.make_executor(self.cluster_params, self._pool)
            self._set_executor(executor, self.cluster_params)
            self._is_snoozing = False
            msg = Message().unsnooze_done("unsnooze done")
            self._event_bus.send(msg)

    def is_snoozing(self):
        return self._is_snoozing

    async def _snooze_check_task(self):
        """
        Periodically check if we need to snooze the executor
        """
        try:
            while True:
                await asyncio.sleep(self._snooze_check_interval)
                if not self._is_snoozing and self._snooze_timeout is not None:
                    since_last_activity = time.monotonic() - self._last_activity
                    if since_last_activity > self._snooze_timeout:
                        await self.snooze()
        except asyncio.CancelledError:
            log.debug("shutting down snooze check task")
            return

    async def make_executor(self, params, pool) -> AsyncAdapter:
        connection = params['connection']
        if connection["type"].lower() == "tcp":
            sync_executor = await sync_to_async(
                DaskJobExecutor.connect,
                pool=self._pool,
                scheduler_uri=connection['address'],
            )
        elif connection["type"].lower() == "local":
            sync_executor = await sync_to_async(
                create_executor,
                pool=self._pool,
                connection=connection,
                local_directory=self.get_local_directory(),
                preload=self.get_preload(),
            )
        else:
            raise ValueError("unknown connection type")
        executor = AsyncAdapter(wrapped=sync_executor, pool=pool)
        return executor

    async def get_executor(self):
        self._update_last_activity()
        await self.unsnooze()
        if self.executor is None:
            # TODO: exception type, conversion into 400 response
            raise RuntimeError("wrong state: executor is None")
        return self.executor

    def have_executor(self):
        return self.executor is not None or self._is_snoozing

    async def get_resource_details(self):
        # memoize the cluster details, if ever we support
        # dynamic resources this will need to change
        if self.cluster_details is None:
            executor = await self.get_executor()
            self.cluster_details = await executor.get_resource_details()
        return self.cluster_details

    async def get_context(self) -> Context:
        await self.unsnooze()
        self._update_last_activity()
        if self.context is None:
            raise RuntimeError("cannot get context, please call `set_executor` before")
        return self.context

    def shutdown(self):
        self._loop.call_soon_threadsafe(self._snooze_task.cancel)
        if self.context is not None:
            self.context.close()

    async def set_executor(self, executor: JobExecutor, params):
        """
        Set the new executor used to run jobs, and the parameters used
        to create/connect.

        After the executor is set, we take "ownership" of it, and ensure
        that it is properly cleaned up in the `shutdown` method.
        """
        self._update_last_activity()
        if self.executor is not None:
            await self.executor.close()
            self.executor = None
        self._set_executor(executor, params)

    def _set_executor(self, executor: JobExecutor, params):
        if self.executor is not None:
            self.executor.ensure_sync().close()
        self.executor = executor
        self.cluster_params = params
        self.context = Context(executor=executor.ensure_sync())
        self._is_snoozing = False  # if we were snoozing before, we no longer are
        try:
            # Exposing the scheduler address allows use of
            # libertem-server to spin up a LT-compatible
            # persistent dask cluster
            log.info(f'Scheduler at {self.context.executor.client.scheduler.address}')
        except AttributeError:
            pass

    def get_cluster_params(self):
        self._update_last_activity()
        return self.cluster_params


class AnalysisState:
    def __init__(self, executor_state: ExecutorState, job_state: 'JobState'):
        self.analyses: dict[str, AnalysisInfo] = {}
        self.results: dict[str, AnalysisResultInfo] = {}
        self.job_state = job_state

    def create(
        self, uuid: str, dataset_uuid: str, analysis_type: str, parameters: AnalysisParameters
    ) -> None:
        assert uuid not in self.analyses
        self.analyses[uuid] = {
            "dataset": dataset_uuid,
            "analysis": uuid,
            "jobs": [],
            "details": {
                "analysisType": analysis_type,
                "parameters": parameters,
            },
        }

    def add_job(self, analysis_id: str, job_id: str) -> None:
        jobs = self.analyses[analysis_id]["jobs"]
        jobs.append(job_id)

    def update(self, uuid: str, analysis_type: str, parameters: AnalysisParameters) -> None:
        self.analyses[uuid]["details"]["parameters"] = parameters
        self.analyses[uuid]["details"]["analysisType"] = analysis_type

    def get(
        self, uuid: str, default: typing.Optional[AnalysisInfo] = None
    ) -> typing.Optional[AnalysisInfo]:
        return self.analyses.get(uuid, default)

    def filter(self, predicate: typing.Callable[[AnalysisInfo], bool]) -> list[AnalysisInfo]:
        return [
            analysis
            for analysis in self.analyses.values()
            if predicate(analysis)
        ]

    async def remove(self, uuid: str) -> bool:
        if uuid not in self.analyses:
            return False
        if uuid in self.results:
            self.remove_results(uuid)
        await self.remove_jobs(uuid)
        del self.analyses[uuid]
        return True

    async def remove_jobs(self, uuid: str) -> None:
        jobs = copy.copy(self.job_state.get_for_analysis_id(uuid))
        for job_id in jobs:
            await self.job_state.remove(job_id)

    def remove_results(self, uuid: str) -> None:
        del self.results[uuid]

    def set_results(
        self,
        analysis_id: str,
        details: AnalysisDetails,
        results: AnalysisResultSet,
        job_id: str,
        udf_results: typing.Optional[UDFResults],
    ) -> None:
        """
        set results: create or update
        """
        self.results[analysis_id] = AnalysisResultInfo(
            copy.deepcopy(details), results, job_id, udf_results
        )

    def have_results(self, analysis_id: str) -> bool:
        return analysis_id in self.results

    def get_results(self, analysis_id: str) -> AnalysisResultInfo:
        return self.results[analysis_id]

    def get_all_results(self) -> typing.ItemsView[str, AnalysisResultInfo]:
        return self.results.items()

    def __getitem__(self, analysis_id: str) -> AnalysisInfo:
        return self.analyses[analysis_id]

    def serialize(self, analysis_id: str) -> AnalysisInfo:
        result = copy.copy(self[analysis_id])
        result["jobs"] = [
            job_id
            for job_id in result["jobs"]
            if not self.job_state.is_cancelled(job_id)
        ]
        return result

    def serialize_all(self) -> list[AnalysisInfo]:
        return [
            self.serialize(analysis_id)
            for analysis_id in self.analyses
        ]


class CompoundAnalysisState:
    def __init__(self, analysis_state: AnalysisState):
        self.analysis_state = analysis_state
        self.analyses: dict[str, CompoundAnalysisInfo] = {}

    def create_or_update(
        self, uuid: str, main_type: str, dataset_id: str, analyses: list[str]
    ) -> bool:
        created = uuid not in self.analyses
        self.analyses[uuid] = {
            "dataset": dataset_id,
            "compoundAnalysis": uuid,
            "details": {
                "mainType": main_type,
                "analyses": analyses,
            }
        }
        return created

    def remove(self, uuid: str) -> None:
        del self.analyses[uuid]

    def __getitem__(self, uuid: str) -> CompoundAnalysisInfo:
        return self.analyses[uuid]

    def filter(
        self, predicate: typing.Callable[[CompoundAnalysisInfo], bool]
    ) -> list[CompoundAnalysisInfo]:
        return [
            ca
            for ca in self.analyses.values()
            if predicate(ca)
        ]

    def serialize(self, uuid: str) -> CompoundAnalysisInfo:
        return self[uuid]

    def serialize_all(self) -> list[CompoundAnalysisInfo]:
        return [
            self.serialize(uuid)
            for uuid in self.analyses
        ]


class DatasetState:
    def __init__(self, executor_state: ExecutorState, analysis_state: AnalysisState,
                 compound_analysis_state: CompoundAnalysisState):
        self.datasets: dict[str, DatasetInfo] = {}
        self.dataset_to_id: dict[DataSet, str] = {}
        self.executor_state = executor_state
        self.analysis_state = analysis_state
        self.compound_analysis_state = compound_analysis_state

    def register(
        self, uuid: str, dataset: DataSet, params: dict, converted: dict
    ):
        assert uuid not in self.datasets
        self.datasets[uuid] = {
            "dataset": dataset,
            "params": params,
            "converted": converted,
        }
        self.dataset_to_id[dataset] = uuid
        return self

    async def serialize(self, dataset_id):
        executor = await self.executor_state.get_executor()
        dataset = self.datasets[dataset_id]
        diag = await executor.run_function(lambda: dataset["dataset"].diagnostics)
        return {
            "id": dataset_id,
            "params": {
                **dataset["params"]["params"],
                "shape": tuple(dataset["dataset"].shape),
            },
            "diagnostics": diag,
        }

    async def serialize_all(self):
        return [
            await self.serialize(dataset_id)
            for dataset_id in self.datasets.keys()
        ]

    def id_for_dataset(self, dataset):
        return self.dataset_to_id[dataset]

    def __getitem__(self, uuid):
        return self.datasets[uuid]["dataset"]

    def __contains__(self, uuid):
        return uuid in self.datasets

    async def verify(self):
        executor = await self.executor_state.get_executor()
        for uuid, params in self.datasets.items():
            dataset = params["dataset"]
            try:
                await executor.run_function(dataset.check_valid)
            except DataSetException:
                await self.remove_dataset(uuid)

    async def remove(self, uuid):
        """
        Stop all jobs and remove dataset state for the dataset identified by `uuid`
        """
        ds = self.datasets[uuid]["dataset"]
        analyses = self.analysis_state.filter(lambda a: a["dataset"] == uuid)
        compound_analyses = self.compound_analysis_state.filter(lambda ca: ca["dataset"] == uuid)
        del self.datasets[uuid]
        del self.dataset_to_id[ds]
        for analysis in analyses:
            await self.analysis_state.remove(analysis["analysis"])
        for ca in compound_analyses:
            self.compound_analysis_state.remove(ca["compoundAnalysis"])


class JobState:
    def __init__(self, executor_state: ExecutorState):
        self.jobs: dict[str, JobInfo] = {}
        self.executor_state = executor_state
        self.jobs_for_dataset = typing.DefaultDict[str, set[str]](lambda: set())
        self.jobs_for_analyses = typing.DefaultDict[str, set[str]](lambda: set())

    def register(self, job_id: str, analysis_id, dataset_id):
        assert job_id not in self.jobs
        self.jobs[job_id] = {
            "id": job_id,
            "analysis": analysis_id,
            "dataset": dataset_id,
        }
        self.jobs_for_dataset[dataset_id].add(job_id)
        self.jobs_for_analyses[analysis_id].add(job_id)
        return self

    async def remove(self, uuid: str) -> bool:
        try:
            executor = await self.executor_state.get_executor()
            await executor.cancel(uuid)
            del self.jobs[uuid]
            for ds, jobs in itertools.chain(self.jobs_for_dataset.items(),
                                            self.jobs_for_analyses.items()):
                if uuid in jobs:
                    jobs.remove(uuid)
            return True
        except KeyError:
            return False

    def get_for_dataset_id(self, dataset_id: str) -> set[str]:
        return self.jobs_for_dataset[dataset_id]

    def get_for_analysis_id(self, analysis_id: str) -> set[str]:
        return self.jobs_for_analyses[analysis_id]

    def __getitem__(self, uuid: str) -> JobInfo:
        return self.jobs[uuid]

    def __contains__(self, uuid: str) -> bool:
        return uuid in self.jobs

    def is_cancelled(self, uuid: str) -> bool:
        return uuid not in self.jobs

    def serialize(self, job_id: str) -> SerializedJobInfo:
        job = self[job_id]
        return {
            "id": job["id"],
            "analysis": job["analysis"],
        }

    def serialize_all(self) -> list[SerializedJobInfo]:
        return [
            self.serialize(job_id)
            for job_id in self.jobs.keys()
        ]


class SharedState:
    def __init__(self, executor_state: "ExecutorState"):
        self.executor_state = executor_state
        self.job_state = JobState(self.executor_state)
        self.analysis_state = AnalysisState(self.executor_state, job_state=self.job_state)
        self.compound_analysis_state = CompoundAnalysisState(self.analysis_state)
        self.dataset_state = DatasetState(
            self.executor_state,
            analysis_state=self.analysis_state,
            compound_analysis_state=self.compound_analysis_state,
        )

    def get_local_cores(self, default: int = 2) -> int:
        cores: typing.Optional[int] = psutil.cpu_count(logical=False)
        if cores is None:
            cores = default
        return cores

    def get_ds_type_info(self, ds_type_id: str):
        from libertem.io.dataset import get_dataset_cls
        cls = get_dataset_cls(ds_type_id)
        ConverterCls = cls.get_msg_converter()
        converter = ConverterCls()
        schema = converter.SCHEMA
        supported_backends = cls.get_supported_io_backends()
        default_backend = cls.get_default_io_backend().id_
        if not supported_backends:
            default_backend = None
        return {
            "schema": schema,
            "default_io_backend": default_backend,
            "supported_io_backends": supported_backends,
        }

    def get_config(self):
        from libertem.io.dataset import filetypes
        detected_devices = devices.detect()
        ds_types = list(filetypes.keys())
        return {
            "version": libertem.__version__,
            "resultFileFormats": ResultFormatRegistry.get_available_formats(),
            "revision": libertem.revision,
            "localCores": self.get_local_cores(),
            "devices": detected_devices,
            "datasetTypes": {
                ds_type_id.upper(): self.get_ds_type_info(ds_type_id)
                for ds_type_id in ds_types
            },
            "cwd": os.getcwd(),
            # '/' works on Windows, too.
            "separator": '/'
        }

    async def create_and_set_executor(self, spec: dict[str, int]):
        """
        Create a new executor from spec, a dict[str, int]
        compatible with the main arguments of cluster_spec().
        Any values not in spec are filled from a call to detect()

        Any existing executor will first closed by the call
        to self.executor_state._set_executor
        """
        from .helpers import create_executor_external  # circular import
        executor, params = create_executor_external(
            spec,
            self.executor_state.get_local_directory(),
            self.executor_state.get_preload(),
        )
        self.executor_state._set_executor(executor, params)



import json

import pytest
import websockets

from utils import assert_msg
from aio_utils import create_connection

pytestmark = [pytest.mark.web_api]


@pytest.mark.asyncio
async def test_detect_failed(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')
        assert initial_msg['datasets'] == []
        assert initial_msg['jobs'] == []

        path = default_raw._path
        detect_url = f"{base_url}/api/datasets/detect/?token={default_token}"

        async with http_client.get(detect_url, params={"path": path}) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DATASET_DETECTION_FAILED', status='error')


@pytest.mark.asyncio
async def test_detect_hdf5(
    hdf5, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')
        assert initial_msg['datasets'] == []
        assert initial_msg['jobs'] == []

        path = hdf5.filename
        detect_url = f"{base_url}/api/datasets/detect/?token={default_token}"

        async with http_client.get(detect_url, params={"path": path}) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DATASET_DETECTED')
            assert "datasetParams" in resp_json
            params = resp_json["datasetParams"]
            assert params['type'] == 'HDF5'
            assert params['path'] == path
            assert params['ds_path'] == "data"



import numpy as np
import pytest

from libertem.executor.inline import InlineJobExecutor
from libertem.udf import UDF
from libertem.udf.sum import SumUDF
from libertem.common.executor import (
    TaskCommHandler, TaskProtocol, WorkerQueue, JobCancelledError,
)
from libertem.common.exceptions import UDFRunCancelled
from libertem.io.dataset.memory import MemoryDataSet


def test_run_each_worker_1():
    def fn1():
        return "some result"

    executor = InlineJobExecutor()

    results = executor.run_each_worker(fn1)
    assert len(results.keys()) == 1
    assert len(results.keys()) == len(executor.get_available_workers())

    k = next(iter(results))
    result0 = results[k]
    assert result0 == "some result"
    assert k == "inline"


class ThreadsPerWorkerUDF(UDF):
    def get_result_buffers(self):
        return {
            'num_threads': self.buffer(kind='nav', dtype=int),
        }

    def process_frame(self, frame):
        assert self.meta.threads_per_worker is not None, \
            "threads_per_worker should be an integer"
        self.results.num_threads[:] = self.meta.threads_per_worker


def test_inline_num_threads(lt_ctx, default_raw):
    threads = 2
    res = lt_ctx.run_udf(
        dataset=default_raw,
        udf=ThreadsPerWorkerUDF()
    )['num_threads']
    assert np.allclose(res, threads)


class CancelledTaskCommHandler(TaskCommHandler):
    def handle_task(self, task: TaskProtocol, queue: WorkerQueue):
        raise JobCancelledError()

    def start(self):
        pass

    def done(self):
        pass


class CancelledMemoryDataSet(MemoryDataSet):
    def get_task_comm_handler(self) -> TaskCommHandler:
        return CancelledTaskCommHandler()


def test_cancellation(lt_ctx, default_raw):
    cancel_ds = CancelledMemoryDataSet(data=np.zeros((16, 16, 16, 16)))

    with pytest.raises(UDFRunCancelled) as ex:
        lt_ctx.run_udf(dataset=cancel_ds, udf=SumUDF())

    assert ex.match(r"^UDF run cancelled after \d+ partitions$")

    # after cancellation, the executor is still usable:
    _ = lt_ctx.run_udf(dataset=default_raw, udf=SumUDF())



import logging
import asyncio

from libertem.common.async_utils import sync_to_async
from libertem.analysis.base import AnalysisResultSet

log = logging.getLogger(__name__)


def log_message(message, exception=False):
    log_fn = log.info
    if exception:
        log_fn = log.exception
    if "job" in message:
        log_fn("message: {} (job={})".format(message["messageType"], message["job"]))
    elif "analysis" in message:
        log_fn("message: {} (analysis={})".format(message["messageType"], message["analysis"]))
    elif "dataset" in message:
        log_fn("message: {} (dataset={})".format(message["messageType"], message["dataset"]))
    elif "msg" in message:
        log_fn("message: %s: %s", message["messageType"], message['msg'])
    else:
        log_fn("message: %s" % message["messageType"])


async def result_images(results: AnalysisResultSet, pool, save_kwargs=None):
    futures = [
        sync_to_async(result.get_image, pool, save_kwargs)
        for result in results
    ]

    images = await asyncio.gather(*futures)
    return images


class CORSMixin:
    pass
    # FIXME: implement these when we want to support CORS later
#    def set_default_headers(self):
#        self.set_header("Access-Control-Allow-Origin", "*")  # XXX FIXME TODO!!!
#        # self.set_header("Access-Control-Allow-Headers", "x-requested-with")
#        self.set_header('Access-Control-Allow-Methods', 'PUT, POST, GET, OPTIONS')
#
#    def options(self, *args):
#        """
#        for CORS pre-flight requests, no body returned
#        """
#        self.set_status(204)
#        self.finish()



import sys
import builtins

import pytest


real_import = builtins.__import__


def monkey_notfound(name, *args, **kwargs):
    if name in ('cupy', ):
        raise ModuleNotFoundError(f"Mocked module not found {name}")
    return real_import(name, *args, **kwargs)


def monkey_importerror(name, *args, **kwargs):
    if name in ('cupy', ):
        raise ImportError(f"Mocked import error {name}")
    return real_import(name, *args, **kwargs)


def monkey_brokencupycuda(name, *args, **kwargs):
    if name in ('cupy.cuda', ):
        raise AttributeError(f"Mocked import error {name}")
    return real_import(name, *args, **kwargs)


def test_detect_error(monkeypatch):
    monkeypatch.delitem(sys.modules, 'cupy', raising=False)
    monkeypatch.delitem(sys.modules, 'cupy.cuda', raising=False)
    monkeypatch.setattr(builtins, '__import__', monkey_importerror)

    # Self test
    with pytest.raises(ImportError):
        import cupy  # NOQA: F401

    monkeypatch.delitem(sys.modules, 'libertem.utils.devices', raising=False)

    with pytest.warns(RuntimeWarning, match="Mocked import error cupy"):
        from libertem.utils.devices import detect
        result = detect()

    assert not result['has_cupy']


def test_detect_notfound(monkeypatch):
    monkeypatch.delitem(sys.modules, 'cupy', raising=False)
    monkeypatch.delitem(sys.modules, 'cupy.cuda', raising=False)
    monkeypatch.setattr(builtins, '__import__', monkey_notfound)

    # Self test
    with pytest.raises(ModuleNotFoundError):
        import cupy  # NOQA: F401

    monkeypatch.delitem(sys.modules, 'libertem.utils.devices', raising=False)

    from libertem.utils.devices import detect
    result = detect()

    assert not result['has_cupy']


def test_detect_nocuda(monkeypatch):
    try:
        import cupy
    except Exception:
        pytest.skip("Importable CuPy required for this test.")
    monkeypatch.delattr(sys.modules['cupy'], 'cuda', raising=False)
    # Self test
    with pytest.raises(Exception):
        cupy.cuda

    monkeypatch.delitem(sys.modules, 'libertem.utils.devices', raising=False)

    with pytest.warns(RuntimeWarning, match="module 'cupy' has no attribute 'cuda'"):
        from libertem.utils.devices import detect
        result = detect()

    assert not result['has_cupy']


def test_detect_broken(monkeypatch):
    try:
        import cupy
    except Exception:
        pytest.skip("Importable CuPy required for this test.")

    # CuPy can throw all kinds of exceptions, depending on what exactly goes wrong
    class FunkyException(Exception):
        pass

    def badfunc(*args, **kwargs):
        raise FunkyException()

    monkeypatch.setattr(cupy, 'array', badfunc)
    monkeypatch.setattr(cupy, 'zeros', badfunc)

    # Self test
    with pytest.raises(FunkyException):
        cupy.array(cupy.zeros(1,))

    monkeypatch.delitem(sys.modules, 'libertem.utils.devices', raising=False)

    from libertem.utils.devices import detect
    with pytest.warns(RuntimeWarning, match='FunkyException'):
        result = detect()

    assert not result['has_cupy']



import time
import copy
from unittest import mock
import contextlib

import pytest
import numpy as np
import sparse
import scipy.sparse
from libertem.common.buffers import BufferWrapper
from libertem.common.executor import (
    TaskProtocol, WorkerQueue, TaskCommHandler,
)
from libertem.io.dataset.base import MMapBackend
from libertem.udf import UDF, UDFRunCancelled
from libertem.udf.sum import SumUDF
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.base import NoOpUDF, MergeAttrMapping
from libertem.api import Context
from libertem.common.exceptions import ExecutorSpecException


def test_ctx_load(lt_ctx, default_raw):
    lt_ctx.load(
        "raw",
        path=default_raw._path,
        nav_shape=(16, 16),
        dtype="float32",
        sig_shape=(128, 128),
    )


def test_run_empty_udf_list(lt_ctx, default_raw):
    ds = lt_ctx.load(
        "raw",
        path=default_raw._path,
        nav_shape=(16, 16),
        dtype="float32",
        sig_shape=(128, 128),
    )
    with pytest.raises(ValueError) as em:
        lt_ctx.run_udf(dataset=ds, udf=[])
    em.match("^empty list of UDFs - nothing to do!$")


def test_run_udf_with_io_backend(lt_ctx, default_raw):
    io_backend = MMapBackend(enable_readahead_hints=True)
    lt_ctx.load(
        "raw",
        path=default_raw._path,
        io_backend=io_backend,
        dtype="float32",
        nav_shape=(16, 16),
        sig_shape=(128, 128),
    )
    res = lt_ctx.run_udf(dataset=default_raw, udf=SumUDF())
    assert np.array(res['intensity']).shape == (128, 128)


# ignore 'no plottable channels' from NoOpUDF:
@pytest.mark.filterwarnings('ignore:no plottable channels:UserWarning')
@pytest.mark.parametrize('progress', (True, False))
@pytest.mark.parametrize('plots', (None, True))
def test_multi_udf(lt_ctx, default_raw, progress, plots):
    udfs = [NoOpUDF(), SumUDF(), SumSigUDF()]
    combined_res = lt_ctx.run_udf(dataset=default_raw, udf=udfs, progress=progress, plots=plots)
    ref_res = (
        lt_ctx.run_udf(dataset=default_raw, udf=udfs[0]),
        lt_ctx.run_udf(dataset=default_raw, udf=udfs[1]),
        lt_ctx.run_udf(dataset=default_raw, udf=udfs[2]),
    )

    assert isinstance(ref_res[0], dict)
    for index, res in enumerate(ref_res):
        for key in res.keys():
            assert np.all(res[key].data == combined_res[index][key].data)


@pytest.mark.filterwarnings('ignore::UserWarning')  # ignore 'no plottable channels' from NoOpUDF
@pytest.mark.parametrize('progress', (True, False))
@pytest.mark.parametrize('plots', (None, True))
@pytest.mark.asyncio
async def test_multi_udf_async(lt_ctx, default_raw, progress, plots):
    udfs = [NoOpUDF(), SumUDF(), SumSigUDF()]
    combined_res = await lt_ctx.run_udf(
        dataset=default_raw, udf=udfs, progress=progress, plots=plots, sync=False
    )
    single_res = (
        await lt_ctx.run_udf(dataset=default_raw, udf=udfs[0], sync=False),
        await lt_ctx.run_udf(dataset=default_raw, udf=udfs[1], sync=False),
        await lt_ctx.run_udf(dataset=default_raw, udf=udfs[2], sync=False),
    )
    ref_res = (
        lt_ctx.run_udf(dataset=default_raw, udf=udfs[0]),
        lt_ctx.run_udf(dataset=default_raw, udf=udfs[1]),
        lt_ctx.run_udf(dataset=default_raw, udf=udfs[2]),
    )
    assert isinstance(ref_res[0], dict)
    for index, res in enumerate(ref_res):
        for key in res.keys():
            assert np.all(res[key].data == combined_res[index][key].data)
            assert np.all(single_res[index][key].data == combined_res[index][key].data)


@pytest.mark.filterwarnings('ignore::UserWarning')  # ignore 'no plottable channels' from NoOpUDF
@pytest.mark.parametrize('progress', (True, False))
@pytest.mark.parametrize('plots', (None, True))
def test_udf_iter(lt_ctx, default_raw, progress, plots):
    udfs = [NoOpUDF(), SumUDF(), SumSigUDF()]
    for res in lt_ctx.run_udf_iter(dataset=default_raw, udf=udfs, progress=progress, plots=plots):
        ref_res = lt_ctx.run_udf(dataset=default_raw, udf=copy.deepcopy(udfs), roi=res.damage.data)
        # first one is empty, skipping
        # kind='sig'
        for key in ref_res[1].keys():
            ref_item = ref_res[1][key].data
            res_item = res.buffers[1][key].data
            assert np.all(ref_item == res_item)
        # kind='nav'
        for key in ref_res[2].keys():
            ref_item = ref_res[2][key].data[res.damage.data]
            res_item = res.buffers[2][key].data[res.damage.data]
            assert np.all(ref_item == res_item)


@pytest.mark.filterwarnings('ignore::UserWarning')  # ignore 'no plottable channels' from NoOpUDF
@pytest.mark.parametrize('progress', (True, False))
@pytest.mark.parametrize('plots', (None, True))
@pytest.mark.asyncio
async def test_udf_iter_async(lt_ctx, default_raw, progress, plots):
    udfs = [NoOpUDF(), SumUDF(), SumSigUDF()]
    async for res in lt_ctx.run_udf_iter(
            dataset=default_raw, udf=udfs, progress=progress, plots=plots, sync=False):
        # Nested execution of UDFs doesn't work, so we take a copy
        ref_res = lt_ctx.run_udf(dataset=default_raw, udf=copy.deepcopy(udfs), roi=res.damage.data)
        # first one is empty, skipping
        # kind='sig'
        for key in ref_res[1].keys():
            ref_item = ref_res[1][key].data
            res_item = res.buffers[1][key].data
            assert np.all(ref_item == res_item)
        # kind='nav'
        for key in ref_res[2].keys():
            ref_item = ref_res[2][key].data[res.damage.data]
            res_item = res.buffers[2][key].data[res.damage.data]
            assert np.all(ref_item == res_item)


@pytest.mark.filterwarnings('ignore::UserWarning')  # ignore 'no plottable channels' from NoOpUDF
@pytest.mark.parametrize(
    'plots', (
        True,
        [[], ['intensity']],
        [[], [('intensity', np.abs), ('intensity', np.sin)]],
        None
    )
)
def test_plots(lt_ctx, default_raw, plots):
    udfs = [NoOpUDF(), SumUDF(), SumSigUDF()]
    if plots is None:

        def test_channel(udf_result, damage):
            result = udf_result['intensity'].data
            if udf_result['intensity'].kind == 'nav':
                res_damage = damage
            else:
                res_damage = True
            return (result, res_damage)

        real_plots = [
            lt_ctx.plot_class(
                dataset=default_raw, udf=udfs[1], channel=test_channel,
                title="Test title"
            ),
            lt_ctx.plot_class(
                dataset=default_raw, udf=udfs[2], channel=test_channel,
                title="Test title"
            )
        ]
        mock_target = real_plots[0]
    else:
        real_plots = plots
        mock_target = lt_ctx.plot_class
    with mock.patch.object(mock_target, 'update') as plot_update:
        with mock.patch.object(mock_target, 'display') as plot_display:
            if plots is None:
                real_plots[0].display()
                real_plots[1].display()
            combined_res = lt_ctx.run_udf(dataset=default_raw, udf=udfs, plots=real_plots)
            plot_update.assert_called()
            plot_display.assert_called()
    ref_res = (
        lt_ctx.run_udf(dataset=default_raw, udf=udfs[0]),
        lt_ctx.run_udf(dataset=default_raw, udf=udfs[1]),
        lt_ctx.run_udf(dataset=default_raw, udf=udfs[2]),
    )

    assert isinstance(ref_res[0], dict)
    for index, res in enumerate(ref_res):
        for key in res.keys():
            assert np.all(res[key].data == combined_res[index][key].data)


@pytest.mark.parametrize('plots', ([['hello']], [[('hello', np.abs)]]))
def test_plots_fail(lt_ctx, default_raw, plots):
    udfs = [NoOpUDF()]
    with pytest.raises(ValueError):
        lt_ctx.run_udf(dataset=default_raw, udf=udfs, plots=plots)


@pytest.mark.parametrize(
    'use_roi', (False, True)
)
def test_display(lt_ctx, default_raw, use_roi):
    if use_roi:
        roi = np.random.choice((True, False), size=default_raw.shape.nav)
    else:
        roi = None
    udf = SumUDF()
    d = lt_ctx.display(dataset=default_raw, udf=udf, roi=roi)
    print(d._repr_html_())


@pytest.mark.parametrize(
    'dtype', (bool, int, float, None)
)
def test_roi_dtype(lt_ctx, default_raw, dtype):
    roi = np.zeros(default_raw.shape.nav, dtype=dtype)
    roi[0, 0] = 1

    ref_roi = np.zeros(default_raw.shape.nav, dtype=bool)
    ref_roi[0, 0] = True

    udf = SumUDF()
    if roi.dtype is not np.dtype(bool):
        match = f"ROI dtype is {roi.dtype}, expected bool. Attempting cast to bool."
        with pytest.warns(UserWarning, match=match):
            res = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)
    else:
        res = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)

    ref = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=ref_roi)

    assert np.all(res['intensity'].raw_data == ref['intensity'].raw_data)


roi_types = (
    sparse.COO,
    sparse.DOK,
    scipy.sparse.coo_matrix,
    scipy.sparse.csr_matrix,
    scipy.sparse.csr_array,
    list,
    tuple,
    int,
)


@pytest.mark.parametrize(
    'roi_type', roi_types
)
def test_allowed_rois(lt_ctx, default_raw, roi_type):
    roi = np.zeros(default_raw.shape.nav, dtype=bool)
    roi[3, 6] = True

    if issubclass(roi_type, sparse.SparseArray):
        roi_input = roi_type.from_numpy(roi)
    elif issubclass(roi_type, scipy.sparse.spmatrix):
        roi_input = roi_type(roi)
    elif hasattr(scipy.sparse, 'sparray') and issubclass(roi_type, scipy.sparse.sparray):
        roi_input = roi_type(roi)
    elif roi_type in (tuple, list):
        roi_input = np.argwhere(roi)
        roi_input = roi_type(roi_type((roi_type(row), True)) for row in roi_input)
    elif roi_type is int:
        roi_input = tuple(np.argwhere(roi).squeeze().tolist())
    else:
        raise ValueError('Unrecognized roi_type')

    udf = SumUDF()
    res = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi_input)
    ref = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)
    assert np.all(res['intensity'].raw_data == ref['intensity'].raw_data)


def test_make_with_inline(inline_executor):
    ctx = Context.make_with('inline')
    assert isinstance(ctx.executor, inline_executor.__class__)


def test_make_with_inline_raises():
    with pytest.raises(ExecutorSpecException):
        Context.make_with('inline', cpus=4)


@pytest.mark.parametrize('n_threads', (None, 4, (2, 3)))
def test_make_with_threads(concurrent_executor, n_threads):
    ctx = Context.make_with('threads', cpus=n_threads)
    assert isinstance(ctx.executor, concurrent_executor.__class__)
    # No way to check number of workers in a concurrent.futures.Executor!


@pytest.mark.parametrize('exec_spec', ('threads', 'inline', 'delayed'))
def test_make_with_raises_gpus_no_support(exec_spec):
    with pytest.raises(ExecutorSpecException):
        Context.make_with(exec_spec, gpus=4)


def test_make_with_unrecognized():
    with pytest.raises(ExecutorSpecException):
        Context.make_with('not_an_executor')


def test_udf_cancellation(default_raw):
    from libertem.common.executor import JobCancelledError
    ctx = Context.make_with('inline')

    # As we don't have an API for explicit cancellation by the user yet, we
    # raise `JobCancelledError` from `process_frame` of the UDF.  In reality, it
    # would be raised either from somewhere inside the UDF runner on explicit
    # cancellation, or from the executor when the data source has run dry.

    class DumbUDF(UDF):
        def get_result_buffers(self):
            return {
                'stuff': self.buffer(kind='nav', dtype='float32'),
            }

        def process_frame(self, frame):
            raise JobCancelledError()

    res_iter = ctx.run_udf_iter(dataset=default_raw, udf=DumbUDF())

    # TODO: support `res_iter.cancel_run()` or similar
    with pytest.raises(UDFRunCancelled) as ex:
        for part in res_iter:
            pass

    assert ex.match(r"^UDF run cancelled after \d+ partitions$")


class DynamicParamsUDF(UDF):
    def __init__(self, latest_index, delay=0.0):
        super().__init__(latest_index=latest_index, delay=delay)

    def get_result_buffers(self) -> dict[str, BufferWrapper]:
        return {
            'index': self.buffer(kind='nav', dtype=int),
            'index_merge': self.buffer(kind='nav', dtype=int),
        }

    def process_partition(self, partition):
        print(f"DynamicParamsUDF {self.params.latest_index}")
        # no need to sleep if we have successfully updated the parameter:
        if self.params.latest_index == 0:
            time.sleep(self.params.delay)
        self.results.index[:] = self.params.latest_index

    def merge(self, dest: MergeAttrMapping, src: MergeAttrMapping):
        dest.index[:] = src.index
        dest.index_merge = self.params.latest_index


def test_dynamic_parameter_update_sync(lt_ctx, default_raw):
    result_iter = lt_ctx.run_udf_iter(
        dataset=default_raw, udf=[DynamicParamsUDF(latest_index=0)]
    )
    with contextlib.closing(result_iter) as result_iter:
        # because this is using the inline executor, we can guarantee
        # that the updated parameters are used for the next partition.
        for idx, part_res in enumerate(result_iter):
            result_iter.update_parameters_experimental([
                {"latest_index": idx + 1}
            ])
            print(idx, part_res.buffers[0]['index'].data)
        # `default_raw` has at least two partitions, so there should be
        # something non-zero in the result:
        assert not np.allclose(part_res.buffers[0]['index'], 0)
        assert not np.allclose(part_res.buffers[0]['index_merge'], 0)


@pytest.mark.parametrize('executor', ['pipelined', 'dask', 'inline', 'concurrent'])
def test_dynamic_parameter_update_integration(
    executor, local_cluster_ctx, pipelined_ctx, concurrent_executor,
):

    # we do this dance to re-use the existing executors, so we don't have to
    # mark this test as slow:
    ctx: Context
    if executor == 'dask':
        ctx = local_cluster_ctx
    elif executor == 'pipelined':
        ctx = pipelined_ctx
    elif executor == 'inline':
        ctx = Context.make_with('inline')
    elif executor == 'concurrent':
        ctx = Context(executor=concurrent_executor)
    else:
        raise ValueError('invalid executor name')

    num_workers = sum(
        w.nthreads for w in ctx.executor.get_available_workers()
    )
    parts = 15
    delay = 0.025

    # this should make it more likely that we wait long enough in the main
    # process for the first task to finish, meaning we get to update the index
    # as soon as possible
    delay_udf = delay / 2
    dataset = ctx.load(
        "memory",
        num_partitions=num_workers * parts,
        datashape=(num_workers * parts, 16, 16),
        sig_dims=2,
    )

    class DelayingCommHandler(TaskCommHandler):
        def handle_task(self, task: TaskProtocol, queue: WorkerQueue):
            # our tests only work if the tasks don't all get submitted at the
            # beginning of the `run_tasks` call - this simulates the live
            # processing scenario
            time.sleep(delay)
    dataset.get_task_comm_handler = lambda: DelayingCommHandler()

    result_iter = ctx.run_udf_iter(
        dataset=dataset, udf=[DynamicParamsUDF(latest_index=0, delay=delay_udf)]
    )
    with contextlib.closing(result_iter) as result_iter:
        for idx, part_res in enumerate(result_iter):
            result_iter.update_parameters_experimental([
                {"latest_index": idx + 1}
            ])
            print(idx, part_res.buffers[0]['index'].data)
        assert not np.allclose(part_res.buffers[0]['index'], 0)
        assert not np.allclose(part_res.buffers[0]['index_merge'], 0)

    # second run, to make sure we don't mess up any state:
    result_iter = ctx.run_udf_iter(
        dataset=dataset, udf=[DynamicParamsUDF(latest_index=0, delay=delay_udf)]
    )
    with contextlib.closing(result_iter) as result_iter:
        for idx, part_res in enumerate(result_iter):
            result_iter.update_parameters_experimental([
                {"latest_index": idx + 1}
            ])
            print(idx, part_res.buffers[0]['index'].data)
        assert not np.allclose(part_res.buffers[0]['index'], 0)
        assert not np.allclose(part_res.buffers[0]['index_merge'], 0)


@pytest.mark.asyncio
async def test_dynamic_parameter_update_async(lt_ctx, default_raw):
    result_iter = lt_ctx.run_udf_iter(
        dataset=default_raw,
        udf=[DynamicParamsUDF(latest_index=0)],
        sync=False,
    )
    try:
        # because this is using the inline executor, we can guarantee
        # that the updated parameters are used for the next partition.
        idx = 0  # no aenumerate in stdlib :(
        async for part_res in result_iter:
            await result_iter.update_parameters_experimental([
                {"latest_index": idx + 1}
            ])
            print(idx, part_res.buffers[0]['index'].data)
            idx += 1
        # `default_raw` has at least two partitions, so there should be
        # something non-zero in the result:
        assert not np.allclose(part_res.buffers[0]['index'], 0)
        assert not np.allclose(part_res.buffers[0]['index_merge'], 0)
    finally:
        await result_iter.aclose()


class DynamicParamsAuxUDF(UDF):
    def __init__(self, latest_index):
        super().__init__(latest_index=latest_index)

    def get_result_buffers(self) -> dict[str, BufferWrapper]:
        return {
            'index': self.buffer(kind='nav', dtype=int),
            'index_merge': self.buffer(kind='nav', dtype=int),
        }

    def process_partition(self, partition):
        print(f"DynamicParamsAuxUDF.process_partition {self.params.latest_index}")
        self.results.index[:] = self.params.latest_index[:, 2]
        # the correct view is set:
        assert np.allclose(
            self.meta.coordinates,
            self.params.latest_index[:, :2],
        )

    def merge(self, dest: MergeAttrMapping, src: MergeAttrMapping):
        print(self.params.latest_index.shape)
        print(f"DynamicParamsAuxUDF.merge: {self.params.latest_index}")
        dest.index[:] = src.index
        dest.index_merge = self.params.latest_index[:, 2]


def test_dynamic_parameter_aux_data(lt_ctx, default_raw):
    def _aux_data(idx):
        coords = np.moveaxis(
            np.indices(default_raw.shape.nav),
            0,
            -1,
        ).astype(np.float32).reshape((-1, 2))
        idx_arr = np.zeros((coords.shape[0], 1), dtype=np.float32)
        idx_arr[:] = idx
        aux = DynamicParamsAuxUDF.aux_data(
            data=np.hstack([coords, idx_arr]),
            kind="nav", dtype=np.float32, extra_shape=(3,)
        )
        return aux
    aux = _aux_data(0)
    result_iter = lt_ctx.run_udf_iter(
        dataset=default_raw, udf=[DynamicParamsAuxUDF(latest_index=aux)]
    )
    with contextlib.closing(result_iter) as result_iter:
        # because this is using the inline executor, we can guarantee
        # that the updated parameters are used for the next partition.
        for idx, part_res in enumerate(result_iter):
            aux = _aux_data(idx + 1)
            result_iter.update_parameters_experimental([
                {"latest_index": aux}
            ])
        # `default_raw` has at least two partitions, so there should be
        # something non-zero in the result:
        assert not np.allclose(part_res.buffers[0]['index'], 0)
        assert not np.allclose(part_res.buffers[0]['index_merge'], 0)


@pytest.mark.asyncio
async def test_res_iter_athrow(lt_ctx: Context, default_raw):
    result_iter = lt_ctx.run_udf_iter(
        dataset=default_raw,
        udf=SumUDF(),
        sync=False,
    )

    with pytest.raises(RuntimeError):
        async for res in result_iter:
            await result_iter.athrow(RuntimeError("stuff"))


def test_res_iter_throw(lt_ctx: Context, default_raw):
    result_iter = lt_ctx.run_udf_iter(
        dataset=default_raw,
        udf=SumUDF(),
    )

    with pytest.raises(RuntimeError):
        for res in result_iter:
            result_iter.throw(RuntimeError("stuff"))



import logging

import numpy as np

from libertem.common.math import prod, count_nonzero
from libertem.udf import UDF


log = logging.getLogger(__name__)


class PickUDF(UDF):
    '''
    Load raw data from ROI

    This UDF is meant for frame picking with a very small ROI, usually a single frame.

    .. versionadded:: 0.4.0

    Examples
    --------
    >>> udf = PickUDF()
    >>> roi = np.zeros(dataset.shape.nav, dtype=bool)
    >>> roi[0] = True
    >>> result = ctx.run_udf(dataset=dataset, udf=udf, roi=roi)
    >>> result["intensity"].raw_data[0].shape
    (32, 32)
    '''
    def __init__(self):
        super().__init__()

    def get_preferred_input_dtype(self):
        ''
        return self.USE_NATIVE_DTYPE

    def get_result_buffers(self):
        ''
        dtype = self.meta.input_dtype
        sigshape = tuple(self.meta.dataset_shape.sig)
        if self.meta.roi is not None:
            navsize = count_nonzero(self.meta.roi)
        else:
            navsize = prod(self.meta.dataset_shape.nav)
        warn_limit = 2**28
        loaded_size = prod(sigshape) * navsize * np.dtype(dtype).itemsize
        if loaded_size > warn_limit:
            log.warning("PickUDF is loading %s bytes, exceeding warning limit %s. "
                "Consider using or implementing an UDF to process data on the worker "
                "nodes instead." % (loaded_size, warn_limit))
        # We are using a "single" buffer since we mostly load single frames. A
        # "sig" buffer would work as well, but would require a transpose to
        # accomodate multiple frames in the last and not first dimension.
        # A "nav" buffer would allocate a NaN-filled buffer for the whole dataset.
        return {
            'intensity': self.buffer(
                kind='single', extra_shape=(navsize, ) + sigshape, dtype=dtype
            )
        }

    def process_tile(self, tile):
        ''
        # We work in flattened nav space with ROI applied
        sl = self.meta.slice.get()
        self.results.intensity[sl] = tile

    def merge(self, dest, src):
        ''
        # We receive full-size buffers from each node that
        # contributes at least one frame and rely on the rest being filled
        # with zeros correctly.
        dest.intensity[:] += src.intensity

    def merge_all(self, ordered_results):
        ''
        intensity_chunks = [b.intensity for b in ordered_results.values()]
        intensity_sum = np.stack(intensity_chunks, axis=0).sum(axis=0)
        return {'intensity': intensity_sum}



import numpy as np

from libertem.udf import UDF
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


class SimpleTestByPartitionUDF(UDF):
    def get_result_buffers(self):
        return {
            'test': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_partition(self, partition):
        # coordinates of frames in two partitions of 8x8x8x8 data
        part_1_coords = [[0, 0], [0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6], [0, 7],
                         [1, 0], [1, 1], [1, 2], [1, 3], [1, 4], [1, 5], [1, 6], [1, 7],
                         [2, 0], [2, 1], [2, 2], [2, 3], [2, 4], [2, 5], [2, 6], [2, 7],
                         [3, 0], [3, 1], [3, 2], [3, 3], [3, 4], [3, 5], [3, 6], [3, 7]]
        part_2_coords = [[4, 0], [4, 1], [4, 2], [4, 3], [4, 4], [4, 5], [4, 6], [4, 7],
                         [5, 0], [5, 1], [5, 2], [5, 3], [5, 4], [5, 5], [5, 6], [5, 7],
                         [6, 0], [6, 1], [6, 2], [6, 3], [6, 4], [6, 5], [6, 6], [6, 7],
                         [7, 0], [7, 1], [7, 2], [7, 3], [7, 4], [7, 5], [7, 6], [7, 7]]
        if self.meta.slice.origin == (0, 0, 0):
            assert np.allclose(self.meta.coordinates, part_1_coords)
        elif self.meta.slice.origin == (32, 0, 0):
            assert np.allclose(self.meta.coordinates, part_2_coords)


def test_tiles_by_partition(lt_ctx):
    data = _mk_random(size=(8, 8, 8, 8), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(32, 8, 8),
                            num_partitions=2, sig_dims=2)

    test = SimpleTestByPartitionUDF()
    lt_ctx.run_udf(dataset=dataset, udf=test)


class SimpleTestByFrameUDF(UDF):
    def get_result_buffers(self):
        return {
            'test': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_frame(self, frame):
        # only two slices, each with one frame
        if self.meta.slice.origin == (0, 0, 0):
            assert np.allclose(self.meta.coordinates, [[0, 0]])
        elif self.meta.slice.origin == (1, 0, 0):
            assert np.allclose(self.meta.coordinates, [[0, 1]])


def test_tiles_by_frame(lt_ctx):
    data = _mk_random(size=(8, 8, 8, 8), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(4, 8, 8), sync_offset=62,
                            num_partitions=2, sig_dims=2)

    test = SimpleTestByFrameUDF()
    lt_ctx.run_udf(dataset=dataset, udf=test)


class SimpleTestByTileZeroSyncOffsetUDF(UDF):
    def get_result_buffers(self):
        return {
            'test': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_tile(self, tile):
        # verifying coordinates of frames of some tile_slices
        if self.meta.slice.origin == (0, 0, 0):
            expected_coords = [[0, 0], [0, 1], [0, 2], [0, 3]]
            assert np.allclose(self.meta.coordinates, expected_coords)
        elif self.meta.slice.origin == (32, 0, 0):
            expected_coords = [[4, 0], [4, 1], [4, 2], [4, 3]]
            assert np.allclose(self.meta.coordinates, expected_coords)
        elif self.meta.slice.origin == (60, 0, 0):
            expected_coords = [[7, 4], [7, 5], [7, 6], [7, 7]]
            assert np.allclose(self.meta.coordinates, expected_coords)


def test_tiles_no_offset(lt_ctx):
    data = _mk_random(size=(8, 8, 8, 8), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(4, 8, 8),
                            num_partitions=2, sig_dims=2)

    test = SimpleTestByTileZeroSyncOffsetUDF()
    lt_ctx.run_udf(dataset=dataset, udf=test)


class SimpleTestByTilePositiveSyncOffsetUDF(UDF):
    def get_result_buffers(self):
        return {
            'test': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_tile(self, tile):
        # only two frames in the tile, which are the last two frames of source data
        expected_coords = [[0, 0], [0, 1]]
        assert np.allclose(self.meta.coordinates, expected_coords)


def test_tiles_positive_offset(lt_ctx):
    data = _mk_random(size=(8, 8, 8, 8), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(4, 8, 8),
                            num_partitions=2, sig_dims=2, sync_offset=62)

    test = SimpleTestByTilePositiveSyncOffsetUDF()
    lt_ctx.run_udf(dataset=dataset, udf=test)


class SimpleTestByTileNegativeSyncOffsetUDF(UDF):
    def get_result_buffers(self):
        return {
            'test': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_tile(self, tile):
        # four frames in the tile, which are the first four frames of source data
        expected_coords = [[7, 6], [7, 7]]
        assert np.allclose(self.meta.coordinates, expected_coords)


def test_tiles_negative_offset(lt_ctx):
    data = _mk_random(size=(8, 8, 8, 8), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(4, 8, 8),
                            num_partitions=2, sig_dims=2, sync_offset=-62)

    test = SimpleTestByTileNegativeSyncOffsetUDF()
    lt_ctx.run_udf(dataset=dataset, udf=test)


class SimpleTestByTileWithROIUDF(UDF):
    def get_result_buffers(self):
        return {
            'test': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_tile(self, tile):
        if self.meta.slice.origin == (0, 0, 0):
            expected_coords = [[0, 0], [0, 7]]
            print("meta:", self.meta.coordinates, "expected:", expected_coords)
            assert np.allclose(self.meta.coordinates, expected_coords)
        elif self.meta.slice.origin == (2, 0, 0):
            expected_coords = [[7, 0], [7, 7]]
            print("meta:", self.meta.coordinates, "expected:", expected_coords)
            assert np.allclose(self.meta.coordinates, expected_coords)


def test_tiles_with_roi(lt_ctx):
    data = _mk_random(size=(8, 8, 8, 8), dtype="float32")
    roi = np.random.choice([False], (8, 8))
    roi[0, 0] = True
    roi[0, 7] = True
    roi[7, 0] = True
    roi[7, 7] = True
    dataset = MemoryDataSet(data=data, tileshape=(4, 8, 8),
                            num_partitions=2, sig_dims=2)

    test = SimpleTestByTileWithROIUDF()
    lt_ctx.run_udf(dataset=dataset, udf=test, roi=roi)


class CoordsInGetTaskData(UDF):
    def get_result_buffers(self):
        return {
            'counter': self.buffer(kind='single', dtype=np.int64),
        }

    def process_tile(self, tile):
        pass

    def get_task_data(self):
        assert self.meta._slice is not None
        assert self.meta.coordinates is not None
        ps = np.zeros((self.meta.coordinates.shape[0],), dtype=bool)
        return {
            'ps': ps,
        }

    def postprocess(self):
        self.results.counter[0] += self.task_data.ps.shape[0]

    def merge(self, dest, src):
        dest.counter += src.counter


def test_coordinates_available(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = lt_ctx.load(
        "memory",
        data=data,
        tileshape=(7, 16, 16),
        num_partitions=2,
        sig_dims=2
    )
    roi = np.random.choice([True, False], dataset.shape.nav)

    udf = CoordsInGetTaskData()
    res = lt_ctx.run_udf(dataset=dataset, udf=udf, roi=roi)
    assert res['counter'].data[0] == np.count_nonzero(roi)



from collections import defaultdict
from typing import TYPE_CHECKING, Callable
from collections.abc import Iterator

if TYPE_CHECKING:
    from .executor import TaskProtocol, ResourceDef


class WorkerSet:
    def __init__(self, workers: list["Worker"]):
        self.workers = workers

    def group_by_host(self) -> list["WorkerSet"]:
        """
        returns a list of `WorkerSet`s, each containing the workers for a single host
        """
        by_host = defaultdict(list)
        for worker in self.workers:
            by_host[worker.host].append(worker)

        result = []
        for host, workers in by_host.items():
            result.append(WorkerSet(workers))
        return result

    def get_by_host(self, host: str) -> "WorkerSet":
        return self.filter(lambda w: w.host == host)

    def example(self):
        """
        get a single representative from this WorkerSet
        """
        return self.workers[0]

    def filter(self, fn: Callable[["Worker"], bool]) -> "WorkerSet":
        return WorkerSet([
            w
            for w in self.workers
            if fn(w)
        ])

    def has_cpu(self) -> "WorkerSet":
        return self.filter(lambda worker: bool(worker.resources.get('CPU', False)))

    def has_cuda(self) -> "WorkerSet":
        return self.filter(lambda worker: bool(worker.resources.get('CUDA', False)))

    def has_threaded_workers(self) -> bool:
        return any(w.nthreads != 1 for w in self.workers)

    def concurrency(self) -> int:
        return sum(w.nthreads for w in self.workers)

    def hosts(self) -> set[str]:
        return {worker.host for worker in self.workers}

    def names(self) -> list[str]:
        return [worker.name for worker in self.workers]

    def extend(self, other: "WorkerSet") -> "WorkerSet":
        return WorkerSet(self.workers + other.workers)

    def __iter__(self) -> Iterator["Worker"]:
        return iter(self.workers)

    def __len__(self) -> int:
        return len(self.workers)

    def __repr__(self) -> str:
        return "<WorkerSet {}>".format(
            self.workers,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, WorkerSet):
            return False
        return self.workers == other.workers


class Worker:
    """
    A reference to a worker process identified by `name` running on `host`.
    """
    def __init__(
        self,
        name: str,
        host: str,
        resources: "ResourceDef",
        nthreads: int
    ):
        self.name = name
        self.host = host
        self.resources = resources
        self.nthreads = nthreads

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<Worker {self.name} on {self.host} with {self.resources}, {self.nthreads} threads>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Worker):
            return False
        return self.name == other.name and self.host == other.host

    def __hash__(self) -> int:
        return hash((self.name, self.host))


def _task_fits_on_worker(task: "TaskProtocol", worker: Worker) -> bool:
    worker_resources = worker.resources
    # all resources of the task must be present in the worker resources:
    for k, v in task.get_resources().items():
        if k not in worker_resources:
            return False
        if v > worker_resources[k]:
            return False
    return True


class Scheduler:
    def __init__(self, all_workers: WorkerSet):
        self.workers = all_workers

    def workers_for_task(self, task: "TaskProtocol") -> WorkerSet:
        """
        Given a task, return a `WorkerSet` that can run said task
        according to the resources requested by the task, and the
        resources defined to be available on the worker.
        """
        return self.workers.filter(lambda worker: _task_fits_on_worker(task, worker))

    def effective_worker_count(self) -> int:
        '''
        Return the effective number of workers for partitioning

        This avoids residual partitions that would degrade performance.
        '''
        cpu_workers = self.workers.has_cpu()
        gpu_workers = self.workers.has_cuda()

        # Mixed case: return only CPU workers or GPU workers, whichever is
        # larger, to not have residual partitions in CPU-only or GPU-only
        # processing. Plus, a GPU worker will spin a CPU at 100 % while running.

        if cpu_workers:
            return max(cpu_workers.concurrency(), gpu_workers.concurrency())
        # GPU-only
        elif gpu_workers:
            return gpu_workers.concurrency()
        else:
            return self.workers.concurrency()



import os

import numpy as np
import pytest
import dask.distributed as dd
import cloudpickle

from libertem.io.dataset.cached import CachedDataSet, LRUCacheStrategy
from libertem.io.dataset.base import TilingScheme
from libertem.common import Shape


@pytest.fixture
def default_cached_ds(tmpdir_factory, default_raw, lt_ctx):
    datadir = tmpdir_factory.mktemp('cached_ds_directory')
    strategy = LRUCacheStrategy(capacity=1024*1024*1024)
    ds = CachedDataSet(
        source_ds=default_raw,
        cache_path=datadir,
        strategy=strategy,
    )
    ds = ds.initialize(executor=lt_ctx.executor)
    yield ds


@pytest.fixture(scope='module', autouse=True)
def client_in_background():
    # A running Dask client can introduce a timing issue
    # between automatic closing of a numpy.memmap object and renaming
    # the underlying file
    with dd.LocalCluster() as cluster:
        client = dd.Client(cluster, set_as_default=False)
        yield
        # to fix "distributed.client - ERROR - Failed to reconnect to scheduler after 10.00 seconds, closing client"  # NOQA
        client.close()


def test_simple(default_cached_ds):
    parts = list(default_cached_ds.get_partitions())
    p = parts[0]

    datadir = default_cached_ds._cache_path
    default_raw = default_cached_ds._source_ds

    tileshape = Shape(
        (16,) + tuple(default_raw.shape.sig),
        sig_dims=default_raw.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_raw.shape,
    )

    print(datadir, os.listdir(datadir))

    print(p)

    t_cache = next(p.get_tiles(tiling_scheme=tiling_scheme))
    t_orig = next(next(default_raw.get_partitions()).get_tiles(tiling_scheme=tiling_scheme))

    print(t_cache)
    print(t_orig)
    assert np.allclose(t_cache.data, t_orig.data)

    for p in default_cached_ds.get_partitions():
        for tile in p.get_tiles(tiling_scheme=tiling_scheme):
            pass


@pytest.mark.slow
def test_with_roi(default_cached_ds):
    roi = np.random.choice(a=[0, 1], size=tuple(default_cached_ds.shape.nav))

    tileshape = Shape(
        (16,) + tuple(default_cached_ds.shape.sig),
        sig_dims=default_cached_ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_cached_ds.shape,
    )

    for p in default_cached_ds.get_partitions():
        for tile in p.get_tiles(tiling_scheme=tiling_scheme, roi=roi):
            pass


@pytest.mark.asyncio
async def test_with_dask_executor(tmpdir_factory, default_raw, dask_executor):
    """
    integration test with the dask executor
    """
    datadir = tmpdir_factory.mktemp('cached_ds_directory')
    strategy = LRUCacheStrategy(capacity=1024*1024*1024)
    ds = CachedDataSet(
        source_ds=default_raw,
        cache_path=datadir,
        strategy=strategy,
    )
    ds = ds.initialize(executor=dask_executor)
    list(ds.get_partitions())  # trigger data locality queries


def test_partition_pickles(default_cached_ds):
    """
    assert that we don't do anything unpickleable in the Partition code!
    """
    p = next(default_cached_ds.get_partitions())
    cloudpickle.loads(cloudpickle.dumps(p))



import os
import warnings
import numpy as np

from libertem.common.math import prod
from libertem.common import Shape
from libertem.common.messageconverter import MessageConverter
from .base import (
    DataSet, DataSetException, DataSetMeta,
    BasePartition, File, FileSet, DirectBackend, IOBackend,
)


class RAWDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/RAWDatasetParams.schema.json",
        "title": "RAWDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "RAW"},
            "path": {"type": "string"},
            "dtype": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
            "io_backend": {
                "enum": IOBackend.get_supported(),
            },
        },
        "required": ["type", "path", "nav_shape", "sig_shape", "dtype"]
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path", "dtype", "nav_shape", "sig_shape"]
        }
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


class RawFile(File):
    pass


class RawFileSet(FileSet):
    pass


class RawFileDataSet(DataSet):
    """
    Read raw data from a single file of raw binary data. This reader assumes the following
    format:

     * only raw data (no file header)
     * frames are stored in C-order without additional frame headers
     * dtype supported by numpy

    Examples
    --------

    >>> ds = ctx.load("raw", path=path_to_raw, nav_shape=(16, 16), sig_shape=(128, 128),
    ...               sync_offset=0, dtype="float32",)

    Parameters
    ----------

    path: str
        Path to the file

    nav_shape: tuple of int
        A n-tuple that specifies the size of the navigation region ((y, x), but
        can also be of length 1 for example for a line scan, or length 3 for
        a data cube, for example)

    sig_shape: tuple of int
        Common case: (height, width); but can be any dimensionality

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start

    dtype: numpy dtype
        The dtype of the data as it is on disk. Can contain endian indicator, for
        example >u2 for big-endian 16bit data.
    """
    def __init__(self, path, dtype, scan_size=None, detector_size=None, enable_direct=False,
                 detector_size_raw=None, crop_detector_to=None, tileshape=None,
                 nav_shape=None, sig_shape=None, sync_offset=0, io_backend=None):
        if enable_direct and io_backend is not None:
            raise ValueError("can't specify io_backend and enable_direct at the same time")
        if enable_direct:
            warnings.warn(
                "enable_direct is deprecated; pass "
                "`io_backend=DirectBackend()` instead",
                FutureWarning
            )
            io_backend = DirectBackend()
        super().__init__(io_backend=io_backend)
        # handle backwards-compatability:
        if tileshape is not None:
            warnings.warn(
                "tileshape argument is ignored and will be removed after 0.6.0",
                FutureWarning
            )
        # FIXME execute deprecation after 0.6.0
        if crop_detector_to is not None:
            warnings.warn("crop_detector_to and detector_size_raw are deprecated, "
                          "and will be removed after version 0.6.0. "
                          "please specify sig_shape instead or use a more "
                          "specific DataSet like EMPAD",
                          FutureWarning)
            if detector_size is not None:
                raise ValueError("cannot specify both detector_size and crop_detector_to")
            if detector_size_raw != crop_detector_to:
                raise ValueError("RawFileDataSet can't crop detector anymore, "
                                 "please use EMPAD DataSet")
            detector_size = crop_detector_to
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset
        # handle backwards-compatability:
        if scan_size is not None:
            warnings.warn(
                "scan_size argument is deprecated. please specify nav_shape instead",
                FutureWarning
            )
            if nav_shape is not None:
                raise ValueError("cannot specify both scan_size and nav_shape")
            self._nav_shape = scan_size
        if detector_size is not None:
            warnings.warn(
                "detector_size argument is deprecated. please specify sig_shape instead",
                FutureWarning
            )
            if sig_shape is not None:
                raise ValueError("cannot specify both detector_size and sig_shape")
            self._sig_shape = detector_size
        if self._nav_shape is None:
            raise TypeError("missing 1 required argument: 'nav_shape'")
        if self._sig_shape is None:
            raise TypeError("missing 1 required argument: 'sig_shape'")
        self._path = path
        self._sig_dims = len(self._sig_shape)
        self._dtype = dtype
        self._filesize = None

    def initialize(self, executor):
        self._filesize = executor.run_function(self._get_filesize)
        if int(prod(self._sig_shape)) > int(self._filesize / np.dtype(self._dtype).itemsize):
            raise DataSetException(
                "sig_shape must be less than size: %s" % (
                    int(self._filesize / np.dtype(self._dtype).itemsize)
                )
            )
        self._image_count = int(
            self._filesize / (
                np.dtype(self._dtype).itemsize * prod(self._sig_shape)
            )
        )
        self._nav_shape_product = int(prod(self._nav_shape))
        self._sync_offset_info = self.get_sync_offset_info()
        shape = Shape(self._nav_shape + self._sig_shape, sig_dims=self._sig_dims)
        self._meta = DataSetMeta(
            shape=shape,
            raw_dtype=np.dtype(self._dtype),
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        return self

    def get_diagnostics(self):
        return [
            {"name": "dtype", "value": str(self._meta.raw_dtype)},
        ]

    def _get_filesize(self):
        return os.stat(self._path).st_size

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    @classmethod
    def get_msg_converter(cls):
        return RAWDatasetParams

    def _get_fileset(self):
        return RawFileSet([
            RawFile(
                path=self._path,
                start_idx=0,
                end_idx=self._image_count,
                sig_shape=self.shape.sig,
                native_dtype=self._meta.raw_dtype,
            )
        ])

    def check_valid(self):
        try:
            fileset = self._get_fileset()
            backend = self.get_io_backend().get_impl()
            with backend.open_files(fileset):
                return True
        except (OSError, ValueError) as e:
            raise DataSetException("invalid dataset: %s" % e)

    def get_cache_key(self):
        return {
            "path": self._path,
            # nav_shape + sig_shape; included because changing nav_shape will change
            # the partition structure and cause errors
            "shape": tuple(self.shape),
            "dtype": str(self.dtype),
            "sync_offset": self._sync_offset,
        }

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield RawPartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )

    def __repr__(self):
        return f"<RawFileDataSet of {self.dtype} shape={self.shape}>"


class RawPartition(BasePartition):
    pass



import typing
import logging
import time

import numpy as np
from matplotlib import colors, cm

from libertem.udf.base import UDFRunner

# Import here for backwards compatibility, refs #1031
from libertem.common.viz import encode_image  # NOQA: F401

if typing.TYPE_CHECKING:
    from libertem.common.buffers import BufferWrapper

logger = logging.getLogger(__name__)


def _get_norm(result, norm_cls=colors.Normalize, vmin=None, vmax=None, damage=None):
    if (vmin is not None) and (vmax is not None):
        return norm_cls(vmin=vmin, vmax=vmax)

    result = result.astype(np.float32)

    if damage is None:
        damage = (result != 0)

    damage = damage & np.isfinite(result)

    if damage.sum() == 0:
        return norm_cls(vmin=1, vmax=1)  # all-NaN or all-zero

    if vmin is None:
        vmin = np.min(result[damage])
    if vmax is None:
        vmax = np.max(result[damage])

    return norm_cls(vmin=vmin, vmax=vmax)


def visualize_simple(result, colormap=None, logarithmic=False, vmin=None, vmax=None, damage=None):
    """
    Normalize and visualize ``result`` with ``colormap`` and return the
    resulting RGBA data as an array.

    Parameters
    ----------
    result : numpy.ndarray
        2d array of intensity values

    colormap : matplotlib colormap or None
        colormap used for visualizing intensity values, defaults to ColormapCubehelix()

    Returns
    -------

    np.array
        A numpy array of shape (Y, X, 4) containing RGBA data, suitable for
        passing to `Image.fromarray` in PIL.
    """
    if logarithmic:
        # Convert to the smallest dtype that supports subtractions
        dtype = np.result_type(result, np.int8)
        result = result.astype(dtype)
        cnorm = colors.LogNorm
        result = result - np.min(result) + 1
    else:
        cnorm = colors.Normalize
    if colormap is None:
        colormap = cm.gist_earth
    norm = _get_norm(result, norm_cls=cnorm, vmin=vmin, vmax=vmax, damage=damage)
    shape = result.shape
    normalized = norm(result.reshape((-1,))).reshape(shape)
    colored = colormap(normalized, bytes=True)
    return colored


def get_plottable_2D_channels(buffers):
    return [
        k
        for k, buf in buffers.items()
        # 2D data, removing axes of size 1
        if len(buf.data.squeeze().shape) == 2
    ]


class Live2DPlot:
    """
    Base plotting class for interactive use. Please see the subclasses for
    concrete details.

    .. versionadded:: 0.7.0

    Parameters
    ----------
    dataset : DataSet
        The dataset on which the UDF will be run. This allows
        to determine the shape of the plots for initialization.
    udf : UDF
        The UDF instance this plot is associated to. This needs to be
        the same instance that is passed to
        :meth:`~libertem.api.Context.run_udf`.
    roi : numpy.ndarray or None
        Region of interest (ROI) that the UDF will
        be run on. This is necessary for UDFs where the `extra_shape`
        parameter of result buffers is a function of the ROI, such as
        :class:`~libertem.udf.raw.PickUDF`.
    channel : misc
        Indicate the channel to be plotted.

        - :code:`None`: The first plottable (2D) channel of the UDF is plotted.
        - :code:`str`: The UDF result buffer name that should be plotted.
        - :code:`tuple(str, function(ndarray) -> ndarray)`: The UDF result buffer name that
          should be plotted together with a function that extracts a plottable result
        - :code:`function(udf_result, damage) -> (ndarray, damage)`: Function that derives a
          plottable 2D ndarray and damage indicator from the full
          UDF results and the processed nav space. See :ref:`plotting` for more details!
    title : str
        The plot title. By default UDF class name and channel name.
    min_delta : float
        Minimum time span in seconds between updates to reduce overheads for slow plotting.
    udfresult : UDFResults, optional
        UDF result to initialize the plot data and determine plot shape. If None (default),
        this is determined using :meth:`~libertem.udf.base.UDFRunner.dry_run` on the dataset,
        UDF and ROI. This parameter allows re-using buffers to avoid unnecessary dry runs.
    """
    def __init__(
            self, dataset, udf, roi=None, channel=None, title=None, min_delta=0, udfresult=None
    ):
        if udfresult is None:
            udfresult = UDFRunner.dry_run([udf], dataset, roi)
        eligible_channels = get_plottable_2D_channels(udfresult.buffers[0])
        if channel is None:
            if not eligible_channels:
                raise ValueError(f"No plottable 2D channel found for {udf.__class__.__name__}")
            channel = eligible_channels[0]
            channel_title = channel

        if callable(channel):
            extract = channel
            channel_title = channel.__name__
            channel = None
        elif isinstance(channel, (tuple, list)):
            channel, func = channel
            if channel not in udfresult.buffers[0]:
                raise ValueError(
                    f"channel {channel} not found, have: {udfresult.buffers[0].keys()}"
                )
            kind = udfresult.buffers[0][channel].kind
            if kind == 'nav':
                def extract(udf_results, damage):
                    return (func(udf_results[channel].data), damage)
            else:
                def extract(udf_results, damage):
                    return (func(udf_results[channel].data), True)

            channel_title = f"{func.__name__}({channel})"
        else:
            extract = None
            if channel not in eligible_channels:
                raise ValueError("channel {} not found or not plottable, have: {!r}".format(
                    channel, eligible_channels
                ))
            channel_title = channel

        self._extract = extract
        self.channel = channel
        if title is None:
            title = f"{udf.__class__.__name__}: {channel_title}"
        self.title = title
        self.data, _ = self.extract(udfresult.buffers[0], udfresult.damage)
        self.udf = udf
        self.last_update = 0
        self.min_delta = min_delta

    def get_udf(self):
        """
        Returns the associated UDF instance
        """
        return self.udf

    def extract(
        self, udf_results: dict[str, "BufferWrapper"], damage: "BufferWrapper"
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Extract plotting data from UDF result.


        Parameters
        ----------

        udf_results : UDF result
            (Partial) UDF result

        damage : BufferWrapper
            :class:`~libertem.common.buffers.BufferWraper` with :code:`kind='nav'`
            and :code:`dtype=bool` that indicates the area of the nav dimension that
            has been processed by the UDF already.

        Returns
        -------
        (numpy.ndarray, damage)
            It returns the data and damage of a UDF result buffer indicated by :code:`channel`
            if that is a string, or a numpy.ndarray and damage derived from the UDF results by
            calling :code:`channel` with this method's arguments if it is callable.
        """
        if self._extract is None:
            buffer = udf_results[self.channel]
            squeezed = buffer.masked_data.squeeze()
            res_damage = buffer.valid_mask.squeeze()
            return (squeezed, res_damage)
        else:
            return self._extract(udf_results, damage)

    def new_data(self, udf_results: dict[str, "BufferWrapper"], damage, force: bool = False):
        """
        This method is called with the raw `udf_results` any time a new
        partition has finished processing.

        The :code:`damage` parameter is filtered to only cover finite
        values of :code:`self.data` and passed to :meth:`self.update`,
        which should then be implemented by a subclass.
        """
        t0 = time.time()
        delta = t0 - self.last_update
        if (not force) and delta < self.min_delta:
            return  # don't update plot if we recently updated
        (self.data, damage) = self.extract(udf_results, damage)
        finite = np.isfinite(self.data)
        # duck typing "array-like" since damage can also be bool
        if hasattr(damage, "shape"):
            start_axis = len(damage.shape)
            stop_axis = len(self.data.shape)
            # Reduce surplus axes, for example RGB plotting #1052
            if stop_axis > start_axis:
                finite = np.all(finite, axis=tuple(range(start_axis, stop_axis)))
        damage = damage & finite
        self.update(damage, force=force)
        self.last_update = time.time()
        logger.debug("%s updated in %.3f seconds", self.__class__.__name__, self.last_update - t0)

    def update(self, damage, force=False):
        """
        Update the plot based on `self.data`.

        Parameters
        ----------
        damage : numpy.ndarray Boolean array with the shape of
            :code:`self.data`. It is :code:`True` for all positions in
            :code:`self.data` that contain finite values and have potentially
            been touched by the UDF. This can be used to extract the correct
            plot range by ignoring invalid buffer portions.

        force : bool Force an update, disabling any throttling mechanisms
        """
        raise NotImplementedError()

    def display(self):
        """
        Show the plot, for example in the current Jupyter cell.
        """
        raise NotImplementedError()


class Dummy2DPlot(Live2DPlot):
    '''
    No-op plot. This is useful for test and example code to not
    attempt displaying :code:`matplotlib` plots in a headless environment
    or during batch operation.
    '''
    def update(self, damage, force=False):
        pass

    def display(self):
        pass
